from __future__ import annotations

import inspect
import json
from contextlib import nullcontext as does_not_raise
from dataclasses import fields
from datetime import datetime
from types import SimpleNamespace

import numpy as np
import pytest
from pandas import DataFrame
from pydantic import Field
from pydantic.dataclasses import dataclass
from pydantic.fields import FieldInfo

from zospy import constants
from zospy.analyses.base import (
    AnalysisData,
    AnalysisMetadata,
    AnalysisResult,
    AnalysisSettings,
    BaseAnalysisWrapper,
    _validated_setter,
    new_analysis,
)
from zospy.analyses.decorators import analysis_settings
from zospy.analyses.parsers.types import ValidatedDataFrame
from zospy.analyses.reports.surface_data import SurfaceDataSettings
from zospy.analyses.systemviewers.base import SystemViewerWrapper


def all_subclasses(cls):
    return set(cls.__subclasses__()).union([s for c in cls.__subclasses__() for s in all_subclasses(c)])


analysis_wrapper_classes = all_subclasses(BaseAnalysisWrapper)
analysis_wrapper_classes.remove(SystemViewerWrapper)


class TestValidatedSetter:
    class MockSettings:
        int_setting: int = 1
        string_setting: str = "a"

    def test_get_existing(self):
        settings = _validated_setter(self.MockSettings())

        assert settings.int_setting == self.MockSettings.int_setting

    def test_get_non_existing(self):
        settings = _validated_setter(self.MockSettings())

        with pytest.raises(AttributeError):
            assert settings.non_existing  # type: ignore

    def test_set_existing(self):
        settings = _validated_setter(self.MockSettings())

        settings.int_setting = 2

        assert settings.int_setting == 2

    def test_set_non_existing(self):
        settings = _validated_setter(self.MockSettings())

        with pytest.raises(AttributeError, match="'MockSettings' object has no attribute 'non_existing'"):
            settings.non_existing = 2  # type: ignore


@dataclass
class MockAnalysisData:
    int_data: int = 1
    string_data: str = "a"


@analysis_settings
class MockAnalysisSettings:
    int_setting: int = Field(default=1, description="An integer setting")
    string_setting: str = Field(default="a", description="A string setting")


class MockAnalysis(BaseAnalysisWrapper[MockAnalysisData, MockAnalysisSettings]):
    TYPE = "MockAnalysis"

    def __init__(
        self,
        *,
        int_setting: int = 1,
        string_setting: str = "a",
        block_remove_temp_files: bool = False,
    ):
        super().__init__(settings_kws=locals())

        self.block_remove_temp_files = block_remove_temp_files

    def _create_analysis(self):
        self._analysis = SimpleNamespace(
            metadata=AnalysisMetadata(DateTime=datetime.now(), LensFile="", LensTitle="", FeatureDescription=""),
            header_data=None,
            messages=[],
            Close=lambda: None,
        )

    def run_analysis(self) -> AnalysisData:
        if self.block_remove_temp_files:
            self._remove_config_file = False
            self._remove_text_output_file = False

        return MockAnalysisData()


MockOpticStudioSystem = type("OpticStudioSystem", (), {})


class TestAnalysisWrapper:
    @staticmethod
    def get_settings_defaults(settings_class):
        result = {}

        for field in fields(settings_class):
            if isinstance(field.default, FieldInfo):
                result[field.name] = field.default.default
            else:
                result[field.name] = field.default

        return result

    def test_get_settings_type(self):
        assert MockAnalysis._settings_type == MockAnalysisSettings  # noqa: SLF001

    def test_settings_type_is_specified(self):
        assert MockAnalysis._settings_type is not AnalysisSettings  # noqa: SLF001

    @pytest.mark.parametrize("cls", analysis_wrapper_classes)
    def test_analyses_correct_analysis_name(self, cls):
        assert cls.TYPE is not None
        assert hasattr(constants.Analysis.AnalysisIDM, cls.TYPE)

    @pytest.mark.parametrize("cls", analysis_wrapper_classes)
    def test_init_all_keyword_only_parameters(self, cls):
        assert all(p.kind.name == "KEYWORD_ONLY" for _, p in inspect.signature(cls).parameters.items())

    @pytest.mark.parametrize("cls", analysis_wrapper_classes)
    def test_init_contains_all_settings(self, cls):
        if cls().settings is None:
            return

        init_signature = inspect.signature(cls.__init__)
        settings_fields = fields(cls().settings)

        assert all(field.name in init_signature.parameters for field in settings_fields)

    @pytest.mark.parametrize("cls", analysis_wrapper_classes)
    def test_analyses_default_values(self, cls):
        if cls().settings is None:
            return

        settings_defaults = self.get_settings_defaults(type(cls().settings))
        init_signature = inspect.signature(cls.__init__)

        for field_name, default_value in settings_defaults.items():
            assert field_name in init_signature.parameters
            assert init_signature.parameters[field_name].default == default_value

    def test_change_settings_from_parameters(self):
        analysis = MockAnalysis(int_setting=2, string_setting="b")

        assert analysis.settings.int_setting == 2
        assert analysis.settings.string_setting == "b"

    def test_change_settings_from_object(self):
        settings = MockAnalysisSettings(int_setting=2, string_setting="b")
        analysis = MockAnalysis.with_settings(settings)

        assert analysis.settings.int_setting == 2
        assert analysis.settings.string_setting == "b"

    def test_settings_object_is_copied(self):
        settings = MockAnalysisSettings(int_setting=2, string_setting="b")
        analysis = MockAnalysis.with_settings(settings)

        assert analysis.settings is not settings
        assert analysis.settings == settings

    def test_update_settings_object(self):
        analysis = MockAnalysis(int_setting=1, string_setting="a")

        analysis.update_settings(settings=MockAnalysisSettings(int_setting=2, string_setting="b"))

        assert analysis.settings.int_setting == 2
        assert analysis.settings.string_setting == "b"

    def test_update_settings_dictionary(self):
        analysis = MockAnalysis(int_setting=1, string_setting="a")

        analysis.update_settings(settings_kws={"int_setting": 2, "string_setting": "b"})

        assert analysis.settings.int_setting == 2
        assert analysis.settings.string_setting == "b"

    def test_update_settings_object_and_dictionary(self):
        analysis = MockAnalysis(int_setting=1, string_setting="a")

        analysis.update_settings(
            settings=MockAnalysisSettings(int_setting=2, string_setting="a"), settings_kws={"string_setting": "b"}
        )

        assert analysis.settings.int_setting == 2
        assert analysis.settings.string_setting == "b"

    def test_update_settings_no_dataclass_raises_type_error(self):
        with pytest.raises(TypeError, match="settings should be a dataclass"):
            MockAnalysis().update_settings(settings=123)

    @pytest.mark.parametrize(
        "field_specification",
        [
            "All",
            0,
            1,
            2,
            3,
        ],
    )
    def test_get_field(self, simple_system, field_specification):
        for x in (-5, 5):
            simple_system.SystemData.Fields.AddField(x, 0, 1.0)

        analysis = new_analysis(simple_system, constants.Analysis.AnalysisIDM.RayFan, settings_first=True)

        if field_specification in ("All", 0):
            analysis.Settings.Field.UseAllFields()
        else:
            analysis.Settings.Field.SetFieldNumber(field_specification)

        assert analysis.get_field() == ("All" if field_specification == 0 else field_specification)

    @pytest.mark.parametrize(
        "analysis_type,field_specification,expectation",
        [
            ("RayFan", "All", does_not_raise()),
            ("RayFan", 0, does_not_raise()),
            ("RayFan", 1, does_not_raise()),
            ("RayFan", 2, does_not_raise()),
            ("RayFan", 3, does_not_raise()),
            ("RayFan", 1.0, pytest.raises(ValueError, match=r"^Field value should be")),
            ("RayFan", 1.5, pytest.raises(ValueError, match=r"^Field value should be")),
            ("RayFan", "Invalid", pytest.raises(ValueError, match=r"^Field value should be")),
            ("RayFan", None, pytest.raises(ValueError, match=r"^Field value should be")),
            ("WavefrontMap", "All", pytest.raises(ValueError, match=r"^Could not set field value to")),
            ("WavefrontMap", 1, does_not_raise()),
        ],
    )
    def test_set_field(self, simple_system, analysis_type, field_specification, expectation):
        for x in (-5, 5):
            simple_system.SystemData.Fields.AddField(x, 0, 1.0)

        analysis = new_analysis(
            simple_system, getattr(constants.Analysis.AnalysisIDM, analysis_type), settings_first=True
        )

        with expectation:
            analysis.set_field(field_specification)
            assert analysis.Settings.Field.GetFieldNumber() == (
                0 if field_specification == "All" else field_specification
            )

    @pytest.mark.parametrize(
        "wavelength_specification",
        [
            "All",
            0,
            1,
            2,
            3,
        ],
    )
    def test_get_wavelength(self, simple_system, wavelength_specification):
        # Use FdC_Visible to have three wavelengths
        simple_system.SystemData.Wavelengths.SelectWavelengthPreset(constants.SystemData.WavelengthPreset.FdC_Visible)

        analysis = new_analysis(simple_system, constants.Analysis.AnalysisIDM.RayFan, settings_first=True)

        if wavelength_specification in ("All", 0):
            analysis.Settings.Wavelength.UseAllWavelengths()
        else:
            analysis.Settings.Wavelength.SetWavelengthNumber(wavelength_specification)

        assert analysis.get_wavelength() == ("All" if wavelength_specification == 0 else wavelength_specification)

    @pytest.mark.parametrize(
        "analysis_type,wavelength_specification,expectation",
        [
            ("RayFan", "All", does_not_raise()),
            ("RayFan", 0, does_not_raise()),
            ("RayFan", 1, does_not_raise()),
            ("RayFan", 2, does_not_raise()),
            ("RayFan", 3, does_not_raise()),
            ("RayFan", 1.0, pytest.raises(ValueError, match=r"^Wavelength value should be")),
            ("RayFan", 1.5, pytest.raises(ValueError, match=r"^Wavelength value should be")),
            ("RayFan", "Invalid", pytest.raises(ValueError, match=r"^Wavelength value should be")),
            ("RayFan", None, pytest.raises(ValueError, match=r"^Wavelength value should be")),
            ("WavefrontMap", "All", pytest.raises(ValueError, match=r"^Could not set wavelength value to")),
            ("WavefrontMap", 1, does_not_raise()),
        ],
    )
    def test_set_wavelength(self, simple_system, analysis_type, wavelength_specification, expectation):
        simple_system.SystemData.Wavelengths.SelectWavelengthPreset(constants.SystemData.WavelengthPreset.FdC_Visible)

        analysis = new_analysis(
            simple_system, getattr(constants.Analysis.AnalysisIDM, analysis_type), settings_first=True
        )

        with expectation:
            analysis.set_wavelength(wavelength_specification)

            assert analysis.Settings.Wavelength.GetWavelengthNumber() == (
                0 if wavelength_specification == "All" else wavelength_specification
            )

    @pytest.mark.parametrize(
        "analysis_type,surface_specification,expectation",
        [
            ("SurfaceCurvature", 0, does_not_raise()),
            ("SurfaceCurvature", 1, does_not_raise()),
            ("SurfaceCurvature", 2, does_not_raise()),
            ("SurfaceCurvature", "Image", pytest.raises(ValueError, match=r"^Could not set surface value to")),
            ("SurfaceCurvature", "Objective", pytest.raises(ValueError, match=r"^Could not set surface value to")),
            ("SurfaceCurvature", 1.5, pytest.raises(ValueError, match=r"^Surface value should be")),
            ("SurfaceCurvature", "Invalid", pytest.raises(ValueError, match=r"^Surface value should be")),
            ("SurfaceCurvature", None, pytest.raises(ValueError, match=r"^Surface value should be")),
            ("WavefrontMap", 1, does_not_raise()),
            ("WavefrontMap", "Image", does_not_raise()),
            ("WavefrontMap", "Objective", pytest.raises(ValueError, match=r"^Could not set surface value to")),
            # TODO Implement test for Objective that should pass when such a test becomes possible
        ],
    )
    def test_set_surface(self, simple_system, analysis_type, surface_specification, expectation):
        analysis = new_analysis(
            simple_system, getattr(constants.Analysis.AnalysisIDM, analysis_type), settings_first=True
        )

        with expectation:
            analysis.set_surface(surface_specification)
            assert analysis.Settings.Surface.GetSurfaceNumber() == (
                0 if surface_specification == "Image" else surface_specification
            )

    @pytest.mark.parametrize(
        "temp_file_type,filename",
        [
            ("text_output_file", "test.txt"),
            ("text_output_file", None),
            ("config_file", "test.CFG"),
            ("config_file", None),
        ],
    )
    def test_create_temp_file(self, temp_file_type, filename, tmp_path, monkeypatch):
        analysis = MockAnalysis(block_remove_temp_files=True)
        monkeypatch.setattr(analysis, f"_needs_{temp_file_type}", True)

        if filename:
            path = tmp_path / filename
            path.touch()
        else:
            path = None

        analysis.run(oss=MockOpticStudioSystem(), **{temp_file_type: path})

        if filename:
            assert getattr(analysis, temp_file_type) == path

        assert getattr(analysis, temp_file_type).exists()
        assert getattr(analysis, temp_file_type).is_file()

    @pytest.mark.parametrize(
        "temp_file_type,filename",
        [
            ("text_output_file", "test.ABC"),
            ("text_output_file", "test.TXT"),
            ("config_file", "test.ABC"),
            ("config_file", "test.cfg"),
        ],
    )
    def test_temp_file_wrong_filename(self, temp_file_type, tmp_path, filename, monkeypatch):
        analysis = MockAnalysis()
        monkeypatch.setattr(analysis, f"_needs_{temp_file_type}", True)

        with pytest.raises(ValueError, match="File path should end with ."):
            analysis.run(oss=MockOpticStudioSystem(), **{temp_file_type: tmp_path / filename})

    @pytest.mark.parametrize(
        "temp_file_type,filename",
        [
            ("text_output_file", "test.txt"),
            ("text_output_file", None),
            ("config_file", "test.CFG"),
            ("config_file", None),
        ],
    )
    def test_remove_temp_file(self, temp_file_type, filename, tmp_path, monkeypatch):
        analysis = MockAnalysis()
        monkeypatch.setattr(analysis, f"_needs_{temp_file_type}", True)

        if filename:
            path = tmp_path / filename
            path.touch()
        else:
            path = None

        analysis.run(oss=MockOpticStudioSystem(), **{temp_file_type: path})

        if path:
            assert path.exists()
        else:
            assert not getattr(analysis, temp_file_type).exists()


class TestAnalysisResultJSONConversion:
    settings = SurfaceDataSettings()

    # Only test for non-dataclass results, because dataclass results are tested separately in the corresponding
    # analysis' tests.
    @pytest.mark.parametrize(
        "result_type,result_value,type_info",
        [
            (DataFrame, DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]}), {"data_type": "dataframe"}),
            (np.ndarray, np.array([1, 2, 3]), {"data_type": "ndarray"}),
        ],
    )
    def test_data_to_json(self, result_type, result_value, type_info):
        result = AnalysisResult[result_type, MockAnalysisSettings](
            data=result_value,
            settings=MockAnalysisSettings(),
            metadata=AnalysisMetadata(datetime.now(), "", "", ""),
            header=None,
            messages=None,
        )

        result_json = result.to_json()
        result_dict = json.loads(result_json)

        assert "__analysis_data__" in result_dict
        assert result_dict["__analysis_data__"] == type_info

    def test_settings_to_json(self):
        result = AnalysisResult[ValidatedDataFrame, MockAnalysisSettings](
            data=DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]}),
            settings=MockAnalysisSettings(),
            metadata=AnalysisMetadata(datetime.now(), "", "", ""),
            header=None,
            messages=None,
        )

        result_json = result.to_json()
        result_dict = json.loads(result_json)

        assert "__analysis_settings__" in result_dict
        assert result_dict["__analysis_settings__"] == {
            "data_type": "zospy_class",
            "name": "MockAnalysisSettings",
            "module": "tests.analyses.test_base",
        }

    @pytest.mark.parametrize(
        "result_type,result_value",
        [
            (ValidatedDataFrame, DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})),
            (np.ndarray, np.array([1, 2, 3])),
        ],
    )
    def test_roundtrip(self, result_type, result_value):
        result = AnalysisResult[result_type, MockAnalysisSettings](
            data=result_value,
            settings=self.settings,
            metadata=AnalysisMetadata(datetime.now(), "", "", ""),
            header=None,
            messages=None,
        )

        result_json = result.to_json()

        result_roundtrip = AnalysisResult.from_json(result_json)
        assert all(result_roundtrip.data == result.data)
        assert result_roundtrip.settings == result.settings
        assert result_roundtrip.metadata == result.metadata
        assert result_roundtrip.header == result.header
        assert result_roundtrip.messages == result.messages



import pytest
from pandas.testing import assert_frame_equal

from tests.helpers import assert_dataclass_equal
from zospy.analyses.wavefront import WavefrontMap, ZernikeStandardCoefficients


class TestWavefrontMap:
    def test_can_run(self, simple_system):
        result = WavefrontMap().run(simple_system)
        assert result.data is not None

    def test_to_json(self, simple_system):
        result = WavefrontMap().run(simple_system)
        assert result.from_json(result.to_json()).to_json() == result.to_json()

    @pytest.mark.parametrize(
        "sampling,use_exit_pupil", [("64x64", True), ("64x64", False), ("128x128", True), ("128x128", False)]
    )
    def test_wavefront_map_returns_correct_result(self, simple_system, sampling, use_exit_pupil, expected_data):
        result = WavefrontMap(sampling=sampling, use_exit_pupil=use_exit_pupil).run(simple_system)

        assert_frame_equal(result.data, expected_data.data)


class TestZernikeStandardCoefficients:
    def test_can_run(self, simple_system):
        result = ZernikeStandardCoefficients().run(simple_system)
        assert result.data is not None

    def test_to_json(self, simple_system):
        result = ZernikeStandardCoefficients().run(simple_system)
        assert result.from_json(result.to_json()).to_json() == result.to_json()

    @pytest.mark.parametrize("sampling,maximum_term", [("64x64", 37), ("128x128", 64)])
    def test_zernike_standard_coefficients_returns_correct_result(
        self, simple_system, sampling, maximum_term, expected_data
    ):
        result = ZernikeStandardCoefficients(sampling=sampling, maximum_term=maximum_term).run(simple_system)

        assert_dataclass_equal(
            result.data,
            expected_data.data,
        )

    @pytest.mark.parametrize("sampling,maximum_term", [("64x64", 37), ("128x128", 64)])
    def test_zernike_standard_coefficients_matches_reference_data(
        self, simple_system, sampling, maximum_term, reference_data
    ):
        result = ZernikeStandardCoefficients(sampling=sampling, maximum_term=maximum_term).run(simple_system)

        assert_dataclass_equal(
            result.data,
            reference_data.data,
        )



"""Polarization transmission analysis."""

from __future__ import annotations

import struct
from typing import Annotated

import pandas as pd
from pydantic import Field

from zospy.analyses.base import BaseAnalysisWrapper
from zospy.analyses.decorators import analysis_result, analysis_settings
from zospy.analyses.parsers.transformers import SimpleField, ZospyTransformer
from zospy.analyses.parsers.types import UnitField, ValidatedDataFrame  # noqa: TCH001
from zospy.api import constants
from zospy.utils import zputils

__all__ = ("PolarizationTransmission", "PolarizationTransmissionSettings")


class PolarizationTransmissionTransformer(ZospyTransformer):
    """Transformer for the output of the Polarization Transmission analysis."""

    def chief_ray_transmissions(self, args) -> SimpleField:
        """Transform the chief ray transmission data to a SimpleField."""
        return SimpleField("Chief ray transmission", list(args))

    def field_transmissions(self, args) -> SimpleField:
        """Transform the field transmission data to a SimpleField."""
        return SimpleField("Field transmission", list(args))

    def transmission_table(self, args) -> SimpleField:
        """Transform the transmission table data to a SimpleField."""
        header, rows = args[0]
        return SimpleField("Transmissions", pd.DataFrame(columns=header, data=rows))


@analysis_result
class FieldTransmission:
    field_pos: UnitField = Field(alias="Field Pos")
    total_transmission: float = Field(alias="Total Transmission")
    transmissions: dict[float, float] = Field(alias="Transmission at")


@analysis_result
class ChiefRayTransmission:
    field_pos: UnitField = Field(alias="Field Pos")
    wavelength: dict[int, UnitField] = Field(alias="Wavelength")
    transmissions: ValidatedDataFrame = Field(alias="Transmissions")


@analysis_result
class PolarizationTransmissionResult:
    """Data for the Polarization Transmission analysis."""

    # x_field, y_field, x_phase, y_phase are only defined if the system is polarized
    x_field: float | None = Field(default=None, alias="X-Field")
    y_field: float | None = Field(default=None, alias="Y-Field")
    x_phase: float | None = Field(default=None, alias="X-Phase")
    y_phase: float | None = Field(default=None, alias="Y-Phase")

    grid_size: str = Field(alias="Grid Size")

    field_transmissions: list[FieldTransmission] = Field(alias="Field transmission")
    chief_ray_transmissions: list[ChiefRayTransmission] = Field(alias="Chief ray transmission")


@analysis_settings
class PolarizationTransmissionSettings:
    """Settings for the polarization transmission analysis.

    Attributes
    ----------
    sampling : str | int
        Sampling grid size. Defaults to "32x32".
    unpolarized : bool
        Use unpolarized light. Defaults to `False`.
    jx : float
        Jones x electric field. Defaults to 1.
    jy : float
        Jones y electric field. Defaults to 0.
    x_phase : float
        Phase of the X component of the Jones electric field in degrees. Defaults to 0.
    y_phase : float
        Phase of the Y component of the Jones electric field in degrees. Defaults to 0.
    """

    sampling: str | Annotated[int, Field(ge=0)] = Field(default="32x32", description="Sampling grid size")
    unpolarized: bool = Field(default=False, description="Use unpolarized light")
    jx: float = Field(default=1, description="Jones electric field vector X component")
    jy: float = Field(default=0, description="Jones electric field vector Y component")
    x_phase: float = Field(default=0, description="Jones electric field vector X phase in degrees")
    y_phase: float = Field(default=0, description="Jones electric field vector Y phase in degrees")


class PolarizationTransmission(
    BaseAnalysisWrapper[PolarizationTransmissionResult, PolarizationTransmissionSettings],
    analysis_type="Transmission",
    needs_config_file=True,
    needs_text_output_file=True,
):
    """Polarization transmission analysis."""

    def __init__(
        self,
        *,
        sampling: str | int = "32x32",
        unpolarized: bool = False,
        jx: float = 1,
        jy: float = 0,
        x_phase: float = 0,
        y_phase: float = 0,
    ):
        """Create a new polarization transmission analysis.

        See Also
        --------
        PolarizationTransmissionSettings : Settings for the polarization transmission analysis.
        """
        super().__init__(settings_kws=locals())

    def run_analysis(self) -> PolarizationTransmissionResult:
        """Run the polarization transmission analysis."""
        settings = self.analysis.Settings
        settings.SaveTo(str(self.config_file))

        settings_bytestring = self.config_file.read_bytes()

        settings_bytearray = bytearray(settings_bytestring)

        sampling_value = constants.process_constant(
            constants.Analysis.SampleSizes, zputils.standardize_sampling(self.settings.sampling)
        ).value__

        # Change settings - all byte indices could only be found via reverse engineering :(
        settings_bytearray[56] = sampling_value
        settings_bytearray[60] = int(self.settings.unpolarized)
        settings_bytearray[24:32] = struct.pack("<d", self.settings.jx)
        settings_bytearray[32:40] = struct.pack("<d", self.settings.jy)
        settings_bytearray[40:48] = struct.pack("<d", self.settings.x_phase)
        settings_bytearray[48:56] = struct.pack("<d", self.settings.y_phase)

        self.config_file.write_bytes(settings_bytearray)

        settings.LoadFrom(str(self.config_file))

        # Run analysis
        self.analysis.ApplyAndWaitForCompletion()

        # Parse results
        return self.parse_output(
            "polarization_transmission",
            transformer=PolarizationTransmissionTransformer,
            result_type=PolarizationTransmissionResult,
        )



"""Polarization pupil map analysis."""

from __future__ import annotations

from typing import Literal, Union

import pandas as pd
from pydantic import Field

from zospy.analyses.base import BaseAnalysisWrapper
from zospy.analyses.decorators import analysis_result, analysis_settings
from zospy.analyses.parsers import ZospyTransformer
from zospy.analyses.parsers.transformers import SimpleField
from zospy.analyses.parsers.types import UnitField, ValidatedDataFrame  # noqa: TCH001
from zospy.api import constants
from zospy.utils.pyutils import xtoa
from zospy.utils.zputils import standardize_sampling

__all__ = ("PolarizationPupilMap", "PolarizationPupilMapSettings")


class PolarizationPupilMapTransformer(ZospyTransformer):
    def pupil_map_table(self, args):
        header, rows = args[0]
        table = pd.DataFrame(columns=header, data=rows)

        return SimpleField("Pupil Map", table)


@analysis_result
class PolarizationPupilMapResult:
    wavelength: float = Field(alias="Wavelength")
    field_pos: UnitField[float] = Field(alias="Field Pos")
    x_field: float = Field(alias="X-Field")
    y_field: float = Field(alias="Y-Field")
    x_phase: float = Field(alias="X-Phase")
    y_phase: float = Field(alias="Y-Phase")
    configs: int = Field(alias="Configs")
    surface: int = Field(alias="Surface")
    transmission: UnitField[float] = Field(alias="Transmission")
    pupil_map: ValidatedDataFrame = Field(alias="Pupil Map")


@analysis_settings
class PolarizationPupilMapSettings:
    """Settings for the polarization pupil map analysis.

    Attributes
    ----------
    jx : float
        Jones x electric field. Defaults to 1.
    jy : float
        Jones y electric field. Defaults to 0.
    x_phase : float
        Phase of the X component of the Jones electric field in degrees. Defaults to 0.
    y_phase : float
        Phase of the Y component of the Jones electric field in degrees. Defaults to 0.
    wavelength : int
        The wavelength number that is to be used. Should be an integer specifying the wavelength number.
        Defaults to 1.
    field : int | str
        The field number that is to be used. Must be an integer specifying the field number. Defaults
        to 1.
    surface : str or int
        The surface that is to be analyzed. Either 'Image', or an integer. Defaults to 'Image'.
    sampling : str or int
        The size of the used grid, either string (e.g. '65x65') or int. The integer will be treated as if obtained from
        zospy.constants.Analysis.SampleSizes_ContrastLoss. Defaults to '11x11'.
    add_configs : str
        The add configs string.
    sub_configs : str
        The subtract configs string.
    """

    jx: float = Field(default=1.0, description="Jones electric field vector X component")
    jy: float = Field(default=0.0, description="Jones electric field vector Y component")
    x_phase: float = Field(default=0.0, description="Jones electric field vector X phase in degrees")
    y_phase: float = Field(default=0.0, description="Jones electric field vector Y phase in degrees")
    wavelength: int = Field(default=1, ge=1, description="Wavelength number")
    field: int = Field(default=1, ge=1, description="Field number")
    surface: Literal["Image"] | int = Field(default="Image", description="Surface to analyze")
    sampling: str | int = Field(default="11x11", description="Sampling grid size")
    add_configs: str = Field(default="", description="Add configurations")
    sub_configs: str = Field(default="", description="Subtract configurations")


class PolarizationPupilMap(
    BaseAnalysisWrapper[Union[PolarizationPupilMapResult, None], PolarizationPupilMapSettings],
    analysis_type="PolarizationPupilMap",
    needs_config_file=True,
    needs_text_output_file=True,
):
    """Polarization pupil map analysis."""

    def __init__(
        self,
        *,
        jx: float = 1.0,
        jy: float = 0.0,
        x_phase: float = 0.0,
        y_phase: float = 0.0,
        wavelength: int = 1,
        field: int = 1,
        surface: Literal["Image"] | int = "Image",
        sampling: str | int = "11x11",
        add_configs: str = "",
        sub_configs: str = "",
    ):
        """Create a new polarization pupil map analysis.

        See Also
        --------
        PolarizationPupilMapSettings : Settings for the polarization pupil map analysis.
        """
        super().__init__(settings_kws=locals())

    def run_analysis(self) -> PolarizationPupilMapResult:
        """Run the polarization pupil map analysis."""
        settings = self.analysis.Settings
        config_file = str(self.config_file)

        # MODIFYSETTINGS are defined in the ZPL help files: The Programming Tab > About the ZPL > Keywords
        settings.SaveTo(config_file)
        settings.ModifySettings(config_file, "PPM_JX", xtoa(self.settings.jx, thousands_separator=None))
        settings.ModifySettings(config_file, "PPM_JY", xtoa(self.settings.jy, thousands_separator=None))
        settings.ModifySettings(config_file, "PPM_PX", xtoa(self.settings.x_phase, thousands_separator=None))
        settings.ModifySettings(config_file, "PPM_PY", xtoa(self.settings.y_phase, thousands_separator=None))
        settings.ModifySettings(config_file, "PPM_WAVE", xtoa(int(self.settings.wavelength), thousands_separator=None))
        settings.ModifySettings(config_file, "PPM_FIELD", xtoa(int(self.settings.field), thousands_separator=None))

        if isinstance(self.settings.surface, str):
            settings.ModifySettings(config_file, "PPM_SURFACE", self.settings.surface)
        else:
            settings.ModifySettings(config_file, "PPM_SURFACE", xtoa(self.settings.surface, thousands_separator=None))

        sampling_value = getattr(
            constants.Analysis.SampleSizes_ContrastLoss, standardize_sampling(self.settings.sampling)
        ).value__
        settings.ModifySettings(config_file, "PPM_SAMP", xtoa(sampling_value - 1, thousands_separator=None))
        settings.ModifySettings(config_file, "PPM_ADDCONFIG", self.settings.add_configs)
        settings.ModifySettings(config_file, "PPM_SUBCONFIG", self.settings.sub_configs)

        settings.LoadFrom(config_file)

        # Run analysis
        self.analysis.ApplyAndWaitForCompletion()

        return self.parse_output(
            "polarization_pupil_map",
            transformer=PolarizationPupilMapTransformer,
            result_type=PolarizationPupilMapResult,
        )



from datetime import datetime

import numpy as np
import pandas as pd
import pytest

from zospy.analyses.old.base import (
    AnalysisMessage,
    AnalysisMetadata,
    AnalysisResult,
    AttrDict,
    OnComplete,
    _AnalysisResultJSONDecoder,
    _AnalysisResultJSONEncoder,
    new_analysis,
)
from zospy.api import constants

pytestmark = pytest.mark.old_analyses


def test_oncomplete():
    assert OnComplete.Close == "Close"
    assert OnComplete.Sustain == "Sustain"
    assert OnComplete.Release == "Release"

    with pytest.raises(ValueError, match="'NonExistentValue' is not a valid OnComplete"):
        OnComplete("NonExistentValue")


class TestToJSON:
    @pytest.fixture(scope="class")
    def example_analysis_result(self):
        metadata = AnalysisMetadata(
            DateTime=datetime.now(),
            LensFile="C:\\Path\\To\\Lens\\File.ZOS",
            LensTitle="LensTitle",
            FeatureDescription="",
        )
        messages = [AnalysisMessage(ErrorCode="EC", Message="message")]

        return AnalysisResult(
            analysistype="DummyAnalysis",
            data=AttrDict(
                a=pd.Series({"b": 1, "c": 2, "d": 3, "e": 4}),
                f=pd.DataFrame({"g": [5.6, 7.8, 9.10, 11.12], "h": ["i", "j", "k", "l"]}),
                m="a useless string",
            ),
            metadata=metadata,
            messages=messages,
        )

    def test_restore_metadata(self, example_analysis_result):
        restored_result = AnalysisResult.from_json(example_analysis_result.to_json())

        assert isinstance(restored_result.MetaData, AnalysisMetadata)
        assert example_analysis_result.MetaData == restored_result.MetaData

    def test_restore_messages(self, example_analysis_result):
        restored_result = AnalysisResult.from_json(example_analysis_result.to_json())

        assert isinstance(restored_result.Messages[0], AnalysisMessage)
        assert example_analysis_result.Messages == restored_result.Messages

    def test_restore_attrdict(self, example_analysis_result):
        restored_result = AnalysisResult.from_json(example_analysis_result.to_json())

        assert isinstance(restored_result.Data, AttrDict)

    def test_restore_pandas_series(self):
        series = pd.Series({"a": 1, "b": 2, "c": 3}, dtype=float)

        restored_series = _AnalysisResultJSONDecoder().decode(_AnalysisResultJSONEncoder().encode(series))

        assert isinstance(restored_series, pd.Series)
        assert all(restored_series == series)

    def test_restore_pandas_multi_index_series(self):
        index = pd.MultiIndex.from_product((["a", "b"], [1, 2]))
        series = pd.Series([123, 456, 345, 678], index=index)

        restored_series = _AnalysisResultJSONDecoder().decode(_AnalysisResultJSONEncoder().encode(series))

        assert isinstance(restored_series, pd.Series)
        assert isinstance(restored_series.index, pd.MultiIndex)
        assert all(restored_series == series)

    def test_restore_pandas_dataframe(self):
        dataframe = pd.DataFrame(
            {
                "int_column": [1, 2, 3, 4],
                "float_column": [1.2, 3.4, 5.6, 7.8],
                "str_column": ["this", "ain't", "string", "theory"],
            }
        )

        restored_dataframe = _AnalysisResultJSONDecoder().decode(_AnalysisResultJSONEncoder().encode(dataframe))

        assert isinstance(restored_dataframe, pd.DataFrame)
        assert all(restored_dataframe == dataframe)

    def test_restore_numpy_ndarray(self):
        array = np.array([1.2345, 2.3456, 3.4567, 4.5678])

        restored_array = _AnalysisResultJSONDecoder().decode(_AnalysisResultJSONEncoder().encode(array))

        assert isinstance(restored_array, np.ndarray)
        assert all(restored_array == array)

    def test_restore_datetime(self):
        test_datetime = datetime(year=2023, month=4, day=17, hour=10, minute=41, second=42, microsecond=43)

        restored_datetime = _AnalysisResultJSONDecoder().decode(_AnalysisResultJSONEncoder().encode(test_datetime))

        assert isinstance(restored_datetime, datetime)
        assert restored_datetime == test_datetime


def test_analysis_wrapper_dir(oss):
    opticstudio_analysis_attributes = [
        "GetSettings",
        "GetResults",
        "IsRunning",
        "Apply",
        "ApplyAndWaitForCompletion",
        "Terminate",
        "WaitForCompletion",
        "Close",
        "Release",
        "ToFile",
    ]

    analysis = new_analysis(oss, analysis_type=constants.Analysis.AnalysisIDM.RayFan)

    # Assert for each attribute separately, but in a single test so `oss` does not need to be recreated for every case.
    for attribute in opticstudio_analysis_attributes:
        assert attribute in dir(analysis)



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

__all__ = (
    "POPBeamTypes",
    "POPDataTypes",
    "POPFiberPositions",
    "POPFiberTypes",
    "POPProjectionTypes",
    "POPScaleTypes",
    "POPShowAsTypes",
    "POPZoomTypes",
)

class POPBeamTypes:
    GaussianWaist: POPBeamTypes = None
    GaussianAngle: POPBeamTypes = None
    GaussianSizeAngle: POPBeamTypes = None
    TopHat: POPBeamTypes = None
    File: POPBeamTypes = None
    DLL: POPBeamTypes = None
    Multimode: POPBeamTypes = None
    AstigmaticGaussian: POPBeamTypes = None

class POPDataTypes:
    Irradiance: POPDataTypes = None
    EXIrradiance: POPDataTypes = None
    EYIrradiance: POPDataTypes = None
    Phase: POPDataTypes = None
    EXPhase: POPDataTypes = None
    EYPhase: POPDataTypes = None
    TransferMagnitude: POPDataTypes = None
    TransferPhase: POPDataTypes = None

class POPFiberPositions:
    ChiefRay: POPFiberPositions = None
    SurfaceVertex: POPFiberPositions = None

class POPFiberTypes:
    GaussianWaist: POPFiberTypes = None
    GaussianAngle: POPFiberTypes = None
    GaussianSizeAngle: POPFiberTypes = None
    TopHat: POPFiberTypes = None
    File: POPFiberTypes = None
    DLL: POPFiberTypes = None
    AstigmaticGaussian: POPFiberTypes = None

class POPProjectionTypes:
    AlongBeam: POPProjectionTypes = None
    AlongNormal: POPProjectionTypes = None
    AlongLocalZ: POPProjectionTypes = None

class POPScaleTypes:
    Linear: POPScaleTypes = None
    Log_Minus_5: POPScaleTypes = None
    Log_Minus_10: POPScaleTypes = None
    Log_Minus_15: POPScaleTypes = None

class POPShowAsTypes:
    Surface: POPShowAsTypes = None
    Contour: POPShowAsTypes = None
    GrayScale: POPShowAsTypes = None
    InverseGrayScale: POPShowAsTypes = None
    FalseColor: POPShowAsTypes = None
    InverseFalseColor: POPShowAsTypes = None
    CrossX: POPShowAsTypes = None
    CrossY: POPShowAsTypes = None
    Encircled: POPShowAsTypes = None
    Ensquared: POPShowAsTypes = None
    EnslittedX: POPShowAsTypes = None
    EnslittedY: POPShowAsTypes = None

class POPZoomTypes:
    NoZoom: POPZoomTypes = None
    X2: POPZoomTypes = None
    X4: POPZoomTypes = None
    X8: POPZoomTypes = None
    X16: POPZoomTypes = None



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations
from . import RMSField, RMSFieldMap

__all__ = (
    "RMSField",
    "RMSFieldMap",
    "FieldDensities",
    "FocusDensities",
    "Method",
    "Orientations",
    "RayDensities",
    "ReferTo",
    "WaveDensities",
)

class FieldDensities:
    FieldDens_5: FieldDensities = None
    FieldDens_10: FieldDensities = None
    FieldDens_15: FieldDensities = None
    FieldDens_20: FieldDensities = None
    FieldDens_25: FieldDensities = None
    FieldDens_30: FieldDensities = None
    FieldDens_35: FieldDensities = None
    FieldDens_40: FieldDensities = None
    FieldDens_45: FieldDensities = None
    FieldDens_50: FieldDensities = None
    FieldDens_55: FieldDensities = None
    FieldDens_60: FieldDensities = None
    FieldDens_65: FieldDensities = None
    FieldDens_70: FieldDensities = None
    FieldDens_75: FieldDensities = None
    FieldDens_80: FieldDensities = None
    FieldDens_85: FieldDensities = None
    FieldDens_90: FieldDensities = None
    FieldDens_95: FieldDensities = None
    FieldDens_100: FieldDensities = None

class FocusDensities:
    FocusDens_5: FocusDensities = None
    FocusDens_10: FocusDensities = None
    FocusDens_15: FocusDensities = None
    FocusDens_20: FocusDensities = None
    FocusDens_25: FocusDensities = None
    FocusDens_30: FocusDensities = None
    FocusDens_35: FocusDensities = None
    FocusDens_40: FocusDensities = None
    FocusDens_45: FocusDensities = None
    FocusDens_50: FocusDensities = None
    FocusDens_55: FocusDensities = None
    FocusDens_60: FocusDensities = None
    FocusDens_65: FocusDensities = None
    FocusDens_70: FocusDensities = None
    FocusDens_75: FocusDensities = None
    FocusDens_80: FocusDensities = None
    FocusDens_85: FocusDensities = None
    FocusDens_90: FocusDensities = None
    FocusDens_95: FocusDensities = None
    FocusDens_100: FocusDensities = None

class Method:
    GaussQuad: Method = None
    RectArray: Method = None

class Orientations:
    Plus_Y: Orientations = None
    Minus_Y: Orientations = None
    Plus_X: Orientations = None
    Minus_X: Orientations = None

class RayDensities:
    RayDens_1: RayDensities = None
    RayDens_32x32: RayDensities = None
    RayDens_64x64: RayDensities = None
    RayDens_2: RayDensities = None
    RayDens_128x128: RayDensities = None
    RayDens_3: RayDensities = None
    RayDens_256x256: RayDensities = None
    RayDens_4: RayDensities = None
    RayDens_5: RayDensities = None
    RayDens_512x512: RayDensities = None
    RayDens_1024x1024: RayDensities = None
    RayDens_6: RayDensities = None
    RayDens_2048x2048: RayDensities = None
    RayDens_7: RayDensities = None
    RayDens_4096x4096: RayDensities = None
    RayDens_8: RayDensities = None
    RayDens_8192x8192: RayDensities = None
    RayDens_9: RayDensities = None
    RayDens_10: RayDensities = None
    RayDens_16384x16384: RayDensities = None
    RayDens_11: RayDensities = None
    RayDens_12: RayDensities = None
    RayDens_13: RayDensities = None
    RayDens_14: RayDensities = None
    RayDens_15: RayDensities = None
    RayDens_16: RayDensities = None
    RayDens_17: RayDensities = None
    RayDens_18: RayDensities = None
    RayDens_19: RayDensities = None
    RayDens_20: RayDensities = None

class ReferTo:
    ChiefRay: ReferTo = None
    Centroid: ReferTo = None

class WaveDensities:
    WaveDens_5: WaveDensities = None
    WaveDens_10: WaveDensities = None
    WaveDens_15: WaveDensities = None
    WaveDens_20: WaveDensities = None
    WaveDens_25: WaveDensities = None
    WaveDens_30: WaveDensities = None
    WaveDens_35: WaveDensities = None
    WaveDens_40: WaveDensities = None
    WaveDens_45: WaveDensities = None
    WaveDens_50: WaveDensities = None
    WaveDens_55: WaveDensities = None
    WaveDens_60: WaveDensities = None
    WaveDens_65: WaveDensities = None
    WaveDens_70: WaveDensities = None
    WaveDens_75: WaveDensities = None
    WaveDens_80: WaveDensities = None
    WaveDens_85: WaveDensities = None
    WaveDens_90: WaveDensities = None
    WaveDens_95: WaveDensities = None
    WaveDens_100: WaveDensities = None



from pathlib import Path

# Reference version of OpticStudio, with which all analysis outputs are compared
REFERENCE_VERSION: str = "25.1.1"

# Folder with configuration files for tests of *_fromcfg analyses
CONFIG_DATA_FOLDER: Path = Path("tests/data/config")

# Folder with reference data for analyses
REFERENCE_DATA_FOLDER: Path = Path("tests/data/reference")



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

__all__ = ("LTEdgeSasmpling", "LTRaySampling", "NSCTraceOptions", "OPDMode", "RayStatus", "RaysType", "ZRDFormatType")

class LTEdgeSasmpling:
    S_1X_Low: LTEdgeSasmpling = None
    S_4X: LTEdgeSasmpling = None
    S_16X: LTEdgeSasmpling = None
    S_64X: LTEdgeSasmpling = None
    S_256X: LTEdgeSasmpling = None

class LTRaySampling:
    S_1X_Low: LTRaySampling = None
    S_4X: LTRaySampling = None
    S_16X: LTRaySampling = None
    S_64X: LTRaySampling = None
    S_256X: LTRaySampling = None
    S_1024X: LTRaySampling = None

class NSCTraceOptions:
    None_: NSCTraceOptions = None
    UsePolarization: NSCTraceOptions = None
    UseSplitting: NSCTraceOptions = None
    UsePolarizationSplitting: NSCTraceOptions = None
    UseScattering: NSCTraceOptions = None
    UsePolarizationScattering: NSCTraceOptions = None
    UseSplittingScattering: NSCTraceOptions = None
    UsePolarizationSplittingScattering: NSCTraceOptions = None

class OPDMode:
    None_: OPDMode = None
    Current: OPDMode = None
    CurrentAndChief: OPDMode = None

class RayStatus:
    Parent: RayStatus = None
    Terminated: RayStatus = None
    Reflected: RayStatus = None
    Transmitted: RayStatus = None
    Scattered: RayStatus = None
    Diffracted: RayStatus = None
    GhostedFrom: RayStatus = None
    DiffractedFrom: RayStatus = None
    ScatteredFrom: RayStatus = None
    RayError: RayStatus = None
    BulkScattered: RayStatus = None
    WaveShifted: RayStatus = None
    TIR: RayStatus = None
    OrdinaryRay: RayStatus = None
    ExtraordinaryRay: RayStatus = None
    WaveShiftPL: RayStatus = None

class RaysType:
    Real: RaysType = None
    Paraxial: RaysType = None

class ZRDFormatType:
    UncompressedFullData: ZRDFormatType = None
    CompressedBasicData: ZRDFormatType = None
    CompressedFullData: ZRDFormatType = None



import pytest
from pandas.testing import assert_frame_equal

from zospy.analyses.surface import Curvature


class TestCurvature:
    def test_can_run(self, simple_system):
        result = Curvature().run(simple_system)
        assert result.data is not None

    def test_to_json(self, simple_system):
        result = Curvature(surface=2).run(simple_system)
        assert result.from_json(result.to_json()).to_json() == result.to_json()

    SKIP_SAGITTAL_CURVATURE = "SagitalCurvature is renamed to SagittalCurvature in OpticStudio 24.1.2 and higher"

    @pytest.mark.parametrize(
        "sampling,data,remove,surface,show_as,contour_format,bfs_criterion,bfs_reverse_direction",
        [
            ("65x65", "TangentialCurvature", None, 2, "Surface", "", "MinimumVolume", False),
            pytest.param(
                "129x129",
                "SagitalCurvature",
                None,
                2,
                "Contour",
                "",
                "MinimumVolume",
                False,
                marks=pytest.mark.skip_for_opticstudio_versions(">=24.1.2", SKIP_SAGITTAL_CURVATURE),
            ),
            pytest.param(
                "129x129",
                "SagitalCurvature",
                "BaseROC",
                2,
                "Contour",
                "0.1",
                "MinimumVolume",
                False,
                marks=pytest.mark.skip_for_opticstudio_versions(">=24.1.2", SKIP_SAGITTAL_CURVATURE),
            ),
            pytest.param(
                "129x129",
                "SagittalCurvature",
                None,
                2,
                "Contour",
                "",
                "MinimumVolume",
                False,
                marks=pytest.mark.skip_for_opticstudio_versions("<24.1.2", SKIP_SAGITTAL_CURVATURE),
            ),
            pytest.param(
                "129x129",
                "SagittalCurvature",
                "BaseROC",
                2,
                "Contour",
                "0.1",
                "MinimumVolume",
                False,
                marks=pytest.mark.skip_for_opticstudio_versions("<24.1.2", SKIP_SAGITTAL_CURVATURE),
            ),
            ("33x33", "X_Curvature", "BestFitSphere", 3, "Contour", "", "MinimumRMS", False),
            ("33x33", "X_Curvature", "BestFitSphere", 3, "Contour", "", "MinimumVolume", False),
            ("33x33", "X_Curvature", "BestFitSphere", 3, "Contour", "0.2", "MinimumVolume", True),
        ],
    )
    def test_curvature_returns_correct_result(
        self,
        simple_system,
        sampling,
        data,
        remove,
        surface,
        show_as,
        contour_format,
        bfs_criterion,
        bfs_reverse_direction,
        expected_data,
    ):
        result = Curvature(
            sampling=sampling,
            data=data,
            remove=remove,
            surface=surface,
            show_as=show_as,
            contour_format=contour_format,
            bfs_criterion=bfs_criterion,
            bfs_reverse_direction=bfs_reverse_direction,
        ).run(simple_system)

        assert_frame_equal(result.data.data, expected_data.data.data)



"""OpticStudio analyses from the Rays and spots category."""

from zospy.analyses.raysandspots.ray_fan import RayFan, RayFanSettings
from zospy.analyses.raysandspots.single_ray_trace import (
    SingleRayTrace,
    SingleRayTraceSettings,
)

__all__ = ("SingleRayTrace", "SingleRayTraceSettings", "RayFan", "RayFanSettings")



from typing import ClassVar

import pytest
from pandas.testing import assert_frame_equal

from zospy.analyses.mtf import FFTThroughFocusMTF, HuygensMTF


class TestFFTThroughFocusMTF:
    def test_can_run(self, simple_system):
        result = FFTThroughFocusMTF().run(simple_system)
        assert result.data is not None

    def test_to_json(self, simple_system):
        result = FFTThroughFocusMTF().run(simple_system)
        assert result.from_json(result.to_json()).to_json() == result.to_json()

    @pytest.mark.parametrize(
        "fieldx,fieldy", [(0, 0), (5.5, 0), (0, 5.5), (5.5, 5.5), (-5.5, 0), (0, -5.5), (-5.5, -5.5)]
    )
    def test_field_parsing(self, fieldx, fieldy, simple_system):
        field1 = simple_system.SystemData.Fields.GetField(1)
        field1.X = fieldx
        field1.Y = fieldy
        result = FFTThroughFocusMTF().run(simple_system)

        assert result.data[0].field_coordinate.value[0] == fieldx
        assert result.data[0].field_coordinate.value[1] == fieldy

    @pytest.mark.parametrize("fields", [[(5.5, -5.5)], [(0, 0), (0.0, 5.5), (5.5, 0.0), (5.5, -5.5)]])
    def test_to_dataframe(self, fields, simple_system):
        field1 = simple_system.SystemData.Fields.GetField(1)
        field1.X = fields[0][0]
        field1.Y = fields[0][1]

        for f in fields[1:]:
            simple_system.SystemData.Fields.AddField(f[0], f[1], 1.0)

        result = FFTThroughFocusMTF().run(simple_system)

        df = result.data.to_dataframe()

        for r in result.data:
            assert_frame_equal(
                df.loc[
                    (df["FieldX"] == r.field_coordinate.value[0]) & (df["FieldY"] == r.field_coordinate.value[1]),
                    r.data.reset_index(drop=False).columns,
                ].reset_index(drop=True),
                r.data.reset_index(drop=False),
            )

    @pytest.mark.parametrize(
        "sampling,delta_focus,frequency,number_of_steps,mtf_type",
        [("64x64", 0.1, 0, 5, "Modulation"), ("128x128", 0.3, 3, 10, "Imaginary")],
    )
    def test_fft_through_focus_mtf_returns_correct_result(
        self, simple_system, sampling, delta_focus, frequency, number_of_steps, mtf_type, expected_data
    ):
        result = FFTThroughFocusMTF(
            sampling=sampling,
            delta_focus=delta_focus,
            frequency=frequency,
            number_of_steps=number_of_steps,
            mtf_type=mtf_type,
        ).run(simple_system)

        for i in range(len(result.data)):
            assert_frame_equal(result.data[i].data, expected_data.data[i].data)

    @pytest.mark.parametrize(
        "sampling,delta_focus,frequency,number_of_steps,mtf_type",
        [("64x64", 0.1, 0, 5, "Modulation"), ("128x128", 0.3, 3, 10, "Imaginary")],
    )
    def test_fft_through_focus_mtf_matches_reference_data(
        self, simple_system, sampling, delta_focus, frequency, number_of_steps, mtf_type, reference_data
    ):
        result = FFTThroughFocusMTF(
            sampling=sampling,
            delta_focus=delta_focus,
            frequency=frequency,
            number_of_steps=number_of_steps,
            mtf_type=mtf_type,
        ).run(simple_system)

        for i in range(len(result.data)):
            assert_frame_equal(result.data[i].data, reference_data.data[i].data)

    _FFT_THROUGH_FOCUS_MTF_MTFTYPE_EXPECTED_RETURN: ClassVar = {
        # The expected return does not match constants.Analysis.Settings.Mtf.MtfTypes for fft_through_focus_mtf
        "Modulation": "5",
        "Real": "6",
        "Imaginary": "7",
        "Phase": "8",
        "SquareWave": "9",
    }

    @pytest.mark.parametrize(
        "mtf_type",
        ["Modulation", "Real", "Imaginary", "Phase", "SquareWave"],
    )
    def test_fft_through_focus_mtf_sets_mtftype_correctly(self, simple_system, mtf_type):
        analysis = FFTThroughFocusMTF(mtf_type=mtf_type)
        analysis.run(simple_system, oncomplete="Sustain")

        assert str(analysis.analysis.Settings.Type) == self._FFT_THROUGH_FOCUS_MTF_MTFTYPE_EXPECTED_RETURN[mtf_type]


class TestHuygensMTF:
    def test_can_run(self, simple_system):
        result = HuygensMTF().run(simple_system)
        assert result.data is not None

    def test_to_json(self, simple_system):
        result = HuygensMTF().run(simple_system)
        assert result.from_json(result.to_json()).to_json() == result.to_json()

    @pytest.mark.parametrize(
        "pupil_sampling,image_sampling,image_delta,mtf_type,maximum_frequency",
        [
            ("64x64", "64x64", 0.0, "Modulation", 150.0),
            ("32x32", "64x64", 1.0, "Modulation", 450.0),
            ("128x128", "128x128", 0.0, "Modulation", 314.5),
            ("32x32", "32x32", 0.0, "Modulation", 150.0),
        ],
    )
    def test_huygens_mtf_returns_correct_result(
        self, simple_system, pupil_sampling, image_sampling, image_delta, mtf_type, maximum_frequency, expected_data
    ):
        result = HuygensMTF(
            pupil_sampling=pupil_sampling,
            image_sampling=image_sampling,
            image_delta=image_delta,
            mtf_type=mtf_type,
            maximum_frequency=maximum_frequency,
        ).run(simple_system)

        assert_frame_equal(result.data, expected_data.data)

    @pytest.mark.parametrize(
        "pupil_sampling,image_sampling,image_delta,mtf_type,maximum_frequency",
        [
            ("64x64", "64x64", 0.0, "Modulation", 150.0),
            ("32x32", "64x64", 1.0, "Modulation", 450.0),
            ("128x128", "128x128", 0.0, "Modulation", 314.5),
            ("32x32", "32x32", 0.0, "Modulation", 150.0),
        ],
    )
    def test_huygens_mtf_matches_reference_data(
        self, simple_system, pupil_sampling, image_sampling, image_delta, mtf_type, maximum_frequency, reference_data
    ):
        result = HuygensMTF(
            pupil_sampling=pupil_sampling,
            image_sampling=image_sampling,
            image_delta=image_delta,
            mtf_type=mtf_type,
            maximum_frequency=maximum_frequency,
        ).run(simple_system)

        assert_frame_equal(result.data, reference_data.data)



import pytest

from zospy.api import constants

# ruff: noqa: ARG002


class TestProcessConstant:
    def test_process_constant_converts_string_to_constant(self, zos):
        processed_constant = constants.process_constant(constants.Analysis.AnalysisIDM, "DetectorViewer")

        assert processed_constant == constants.Analysis.AnalysisIDM.DetectorViewer

    def test_process_constant_converts_constant_to_constant(self, zos):
        processed_constant = constants.process_constant(
            constants.SystemData.ZemaxApertureType, constants.SystemData.ZemaxApertureType.FloatByStopSize
        )

        assert processed_constant == constants.SystemData.ZemaxApertureType.FloatByStopSize

    @pytest.mark.parametrize("none", [None, "None", "None_"])
    def test_process_constant_converts_none_to_constant(self, none, zos):
        processed_constant = constants.process_constant(constants.Editors.SolveType, none)

        assert processed_constant == constants.Editors.SolveType.None_

    def test_process_constant_incorrect_string_raises_valueerror(self, zos):
        with pytest.raises(ValueError, match="Constant ZOSAPI_Mode does not contain value"):
            constants.process_constant(constants.ZOSAPI_Mode, "NonExistentMode")

    def test_process_constant_incorrect_constant_raises_valueerror(self, zos):
        with pytest.raises(ValueError, match="Constant SettingsDataType does not contain value"):
            constants.process_constant(constants.Common.SettingsDataType, constants.SystemData.FieldColumn.Comment)



from tests.helpers import assert_dataclass_equal
from zospy.analyses.reports import CardinalPoints, SurfaceData, SystemData


class TestSurfaceData:
    def test_can_run(self, simple_system):
        result = SurfaceData(surface=2).run(simple_system)
        assert result.data is not None

    def test_to_json(self, simple_system):
        result = SurfaceData(surface=2).run(simple_system)
        assert result.from_json(result.to_json()).to_json() == result.to_json()

    def test_surface_data_returns_correct_result(self, simple_system, expected_data):
        result = SurfaceData(surface=2).run(simple_system)

        assert_dataclass_equal(result.data, expected_data.data, ignore_fields=["date", "file"])

    def test_surface_data_matches_reference_data(self, simple_system, reference_data):
        result = SurfaceData(surface=2).run(simple_system)

        assert_dataclass_equal(result.data, reference_data.data, ignore_fields=["date", "file"])


class TestSystemData:
    def test_can_run(self, simple_system):
        result = SystemData().run(simple_system)
        assert result.data is not None

    def test_to_json(self, simple_system):
        result = SystemData().run(simple_system)
        assert result.from_json(result.to_json()).to_json() == result.to_json()

    def test_system_data_returns_correct_result(self, simple_system, expected_data):
        result = SystemData().run(simple_system)

        assert_dataclass_equal(result.data, expected_data.data)

    def test_system_data_matches_reference_data(self, simple_system, reference_data):
        result = SystemData().run(simple_system)

        assert_dataclass_equal(result.data, reference_data.data)


class TestCardinalPoints:
    def test_can_run(self, simple_system):
        result = CardinalPoints(surface_1=2, surface_2=3).run(simple_system)
        assert result.data is not None

    def test_to_json(self, simple_system):
        result = CardinalPoints(surface_1=2, surface_2=3).run(simple_system)
        assert result.from_json(result.to_json()).to_json() == result.to_json()

    def test_cardinal_points_returns_correct_result(self, simple_system, expected_data):
        result = CardinalPoints(surface_1=2, surface_2=3).run(simple_system)

        assert_dataclass_equal(result.data, expected_data.data)

    def test_cardinal_points_matches_reference_data(self, simple_system, reference_data):
        result = CardinalPoints(surface_1=2, surface_2=3).run(simple_system)

        assert_dataclass_equal(result.data, reference_data.data)



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

from zospy.api._ZOSAPI.Common import ZemaxColor
from zospy.api._ZOSAPI.Editors import IEditor, IEditorCell, IEditorRow

__all__ = ("IMeritFunctionEditor", "IMFERow", "MeritColumn", "MeritOperandType")

class IMeritFunctionEditor(IEditor):
    @property
    def RowToOperandOffset(self) -> int: ...
    @property
    def NumberOfOperands(self) -> int: ...
    @property
    def FirstColumn(self) -> MeritColumn: ...
    @property
    def LastColumn(self) -> MeritColumn: ...
    @property
    def MeritFunctionDirectory(self) -> str: ...
    @property
    def NSCOptimizationWizard(self) -> INSCOptimizationWizard: ...
    @property
    def SEQOptimizationWizard(self) -> ISEQOptimizationWizard: ...
    @property
    def NSCBitmapWizard(self) -> INSCBitmapWizard: ...
    @property
    def NSCRoadwayLightingWizard(self) -> INSCRoadwayLightingWizard: ...
    @property
    def SEQOptimizationWizard2(self) -> ISEQOptimizationWizard2: ...
    def AddOperand(self) -> IMFERow: ...
    def CalculateMeritFunction(self) -> float: ...
    def CopyOperands(self, fromOperandNumber: int, NumberOfOperands: int, toOperandNumber: int) -> int: ...
    def CopyOperandsFrom(
        self, fromEditor: IMeritFunctionEditor, fromOperandNumber: int, NumberOfOperands: int, toOperandNumber: int
    ) -> int: ...
    def GetMeritFunctionFiles(self) -> list[str]: ...
    def GetOperandAt(self, OperandNumber: int) -> IMFERow: ...
    def GetOperandValue(
        self,
        type: MeritOperandType,
        srf: int,
        wave: int,
        Hx: float,
        Hy: float,
        Px: float,
        Py: float,
        Ex: float,
        Ey: float,
    ) -> float: ...
    def HideMFE(self) -> None: ...
    def InsertMeritFunction(self, fileName: str, OperandNumber: int) -> int: ...
    def InsertNewOperandAt(self, OperandNumber: int) -> IMFERow: ...
    def LoadMeritFunction(self, fileName: str) -> None: ...
    def RemoveOperandAt(self, OperandNumber: int) -> bool: ...
    def RemoveOperandsAt(self, OperandNumber: int, numOperands: int) -> int: ...
    def SaveMeritFunction(self, fileName: str) -> None: ...
    def ShowMFE(self) -> bool: ...

class IMFERow(IEditorRow):
    @property
    def IsActive(self) -> bool: ...
    @property
    def OperandNumber(self) -> int: ...
    @property
    def TypeName(self) -> str: ...
    @property
    def Type(self) -> MeritOperandType: ...
    @property
    def RowColor(self) -> ZemaxColor: ...
    @RowColor.setter
    def RowColor(self, value: ZemaxColor) -> None: ...
    @property
    def Target(self) -> float: ...
    @Target.setter
    def Target(self, value: float) -> None: ...
    @property
    def TargetCell(self) -> IEditorCell: ...
    @property
    def Weight(self) -> float: ...
    @Weight.setter
    def Weight(self, value: float) -> None: ...
    @property
    def WeightCell(self) -> IEditorCell: ...
    @property
    def Value(self) -> float: ...
    @property
    def ValueCell(self) -> IEditorCell: ...
    @property
    def Contribution(self) -> float: ...
    @property
    def ContributionCell(self) -> IEditorCell: ...
    def AvailableOperandTypes(self) -> list[MeritOperandType]: ...
    def ChangeType(self, type: MeritOperandType) -> bool: ...
    def GetOperandCell(self, Col: MeritColumn) -> IEditorCell: ...

class MeritColumn:
    Comment = 1
    Param1 = 2
    Param2 = 3
    Param3 = 4
    Param4 = 5
    Param5 = 6
    Param6 = 7
    Param7 = 8
    Param8 = 9
    Target = 10
    Weight = 11
    Value = 12
    Contrib = 13

class MeritOperandType:
    ACOS = 0
    ABSO = 1
    CVVA = 2
    DENC = 3
    DENF = 4
    DIFF = 5
    AMAG = 6
    ANAC = 7
    ANAR = 8
    ANAX = 9
    ANAY = 10
    ANCX = 11
    ANCY = 12
    ASIN = 13
    ASTI = 14
    ATAN = 15
    AXCL = 16
    BIOC = 17
    CTVA = 18
    CVGT = 19
    CVLT = 20
    CVOL = 21
    DIMX = 22
    DISC = 23
    DISG = 24
    DIST = 25
    DIVI = 26
    DLTN = 27
    DMFS = 28
    DMGT = 29
    DMLT = 30
    DMVA = 31
    DXDX = 32
    DXDY = 33
    DYDX = 34
    DYDY = 35
    EFFL = 36
    EFLX = 37
    EFLY = 38
    ENDX = 39
    ENPP = 40
    EPDI = 41
    EQUA = 42
    ETGT = 43
    ETLT = 44
    ETVA = 45
    EXPP = 46
    FCGS = 47
    FCGT = 48
    FCUR = 49
    FICL = 50
    FICP = 51
    FOUC = 52
    GBPD = 53
    GBPR = 54
    GBPS = 55
    GBPW = 56
    GBPP = 57
    GBSD = 58
    P1GT = 59
    P1LT = 60
    P1VA = 61
    P2GT = 62
    P2LT = 63
    P2VA = 64
    P3GT = 65
    P3LT = 66
    P3VA = 67
    P4GT = 68
    P4LT = 69
    P4VA = 70
    P5GT = 71
    P5LT = 72
    P5VA = 73
    P6GT = 74
    P6LT = 75
    P6VA = 76
    P7GT = 77
    P7LT = 78
    P7VA = 79
    P8GT = 80
    P8LT = 81
    P8VA = 82
    PANA = 83
    PANB = 84
    PANC = 85
    PARA = 86
    PARB = 87
    PARC = 88
    PARR = 89
    PARX = 90
    PARY = 91
    PARZ = 92
    PATX = 93
    PATY = 94
    PETC = 95
    PETZ = 96
    PIMH = 97
    PLEN = 98
    PMAG = 99
    POWR = 100
    PRIM = 101
    PROD = 102
    PMGT = 103
    PMLT = 104
    PMVA = 105
    POPD = 106
    QOAC = 107
    QSUM = 108
    RAGA = 109
    RAGB = 110
    RAGC = 111
    RAGX = 112
    RAGY = 113
    RAGZ = 114
    RAED = 115
    RAEN = 116
    RAID = 117
    RAIN = 118
    RANG = 119
    REAA = 120
    REAB = 121
    REAC = 122
    REAR = 123
    REAX = 124
    REAY = 125
    REAZ = 126
    RELI = 127
    GBSR = 128
    GBSS = 129
    TGTH = 130
    TMAS = 131
    TOTR = 132
    TRAC = 133
    TRAD = 134
    TRAE = 135
    TRAI = 136
    TRAN = 137
    TRAR = 138
    TRAX = 139
    TRAY = 140
    TRCX = 141
    TRCY = 142
    TTGT = 143
    TTHI = 144
    TTLT = 145
    GBSW = 146
    GBSP = 147
    GCOS = 148
    GENC = 149
    GENF = 150
    GLCA = 151
    GLCB = 152
    GLCC = 153
    GLCX = 154
    GLCY = 155
    GLCZ = 156
    GMTA = 157
    GMTS = 158
    GMTT = 159
    GOTO = 160
    GPIM = 161
    GRMN = 162
    GRMX = 163
    GTCE = 164
    HACG = 165
    HHCN = 166
    I1GT = 167
    BIOD = 168
    BIPF = 169
    BLNK = 170
    BSER = 171
    CENX = 172
    CENY = 173
    CMFV = 174
    CMGT = 175
    CMLT = 176
    CMVA = 177
    CODA = 178
    COGT = 179
    COLT = 180
    COMA = 181
    CONF = 182
    CONS = 183
    COSI = 184
    COVA = 185
    CTGT = 186
    CTLT = 187
    I2GT = 188
    I3GT = 189
    I4GT = 190
    I5GT = 191
    I6GT = 192
    I1LT = 193
    I2LT = 194
    I3LT = 195
    I4LT = 196
    I5LT = 197
    I6LT = 198
    I1VA = 199
    I2VA = 200
    I3VA = 201
    I4VA = 202
    I5VA = 203
    I6VA = 204
    IMAE = 205
    IMSF = 206
    INDX = 207
    ISFN = 208
    ISNA = 209
    LACL = 210
    LINV = 211
    LOGE = 212
    LOGT = 213
    LONA = 214
    LPTD = 215
    MAXX = 216
    MCOG = 217
    MCOL = 218
    MCOV = 219
    MINN = 220
    MNAB = 221
    MNCA = 222
    MNCG = 223
    MNCT = 224
    MNCV = 225
    MNDT = 226
    MNEA = 227
    MNEG = 228
    MNET = 229
    MNIN = 230
    MNPD = 231
    MNSD = 232
    MSWA = 233
    MSWS = 234
    MSWT = 235
    MTFA = 236
    MTFS = 237
    MTFT = 238
    MXAB = 239
    MXCA = 240
    MXCG = 241
    MXCT = 242
    MXCV = 243
    MXDT = 244
    MXEA = 245
    MXEG = 246
    MXET = 247
    MXIN = 248
    MXPD = 249
    MXSD = 250
    NPGT = 251
    NPLT = 252
    NPVA = 253
    NPXG = 254
    NPXL = 255
    NPXV = 256
    NPYG = 257
    NPYL = 258
    NPYV = 259
    NPZG = 260
    NPZL = 261
    NPZV = 262
    NSDC = 263
    NSDD = 264
    NSRA = 265
    NSTR = 266
    NTXG = 267
    NTXL = 268
    NTXV = 269
    NTYG = 270
    NTYL = 271
    NTYV = 272
    NTZG = 273
    NTZL = 274
    NTZV = 275
    OSCD = 276
    OBSN = 277
    OOFF = 278
    OPDC = 279
    OPDM = 280
    OPDX = 281
    OPGT = 282
    OPLT = 283
    OPVA = 284
    OPTH = 285
    OSUM = 286
    RENA = 287
    RENB = 288
    RENC = 289
    RETX = 290
    RETY = 291
    RGLA = 292
    RSCH = 293
    RSCE = 294
    RSRE = 295
    RSRH = 296
    RWCH = 297
    RWCE = 298
    RWRE = 299
    RWRH = 300
    SAGX = 301
    SAGY = 302
    SFNO = 303
    SINE = 304
    SKIS = 305
    SKIN = 306
    SPCH = 307
    SPHA = 308
    SQRT = 309
    SSAG = 310
    SUMM = 311
    SVIG = 312
    TANG = 313
    TFNO = 314
    WFNO = 315
    TTVA = 316
    UDOP = 317
    UDOC = 318
    USYM = 319
    VOLU = 320
    WLEN = 321
    XDVA = 322
    XDGT = 323
    XDLT = 324
    XENC = 325
    XENF = 326
    XNEA = 327
    XNEG = 328
    XNET = 329
    XXEA = 330
    XXEG = 331
    XXET = 332
    YNIP = 333
    ZERN = 334
    ZPLM = 335
    ZTHI = 336
    TOLR = 337
    FTGT = 338
    FTLT = 339
    GLCR = 340
    EFNO = 341
    DIVB = 342
    PROB = 343
    TCGT = 344
    TCLT = 345
    TCVA = 346
    NORX = 347
    NORY = 348
    NORZ = 349
    NORD = 350
    COSA = 351
    MTHA = 352
    MTHS = 353
    MTHT = 354
    BFSD = 355
    EXPD = 356
    CVIG = 357
    GPSX = 358
    GPSY = 359
    GPRX = 360
    GPRY = 361
    GPRT = 362
    POWF = 363
    STHI = 364
    CNPX = 365
    CNPY = 366
    CNAX = 367
    CNAY = 368
    STRH = 369
    CIGT = 370
    CILT = 371
    CIVA = 372
    CEGT = 373
    CELT = 374
    CEVA = 375
    NSST = 376
    ABCD = 377
    DISA = 378
    POWP = 379
    RECI = 380
    NSRM = 381
    FDMO = 382
    FDRE = 383
    FREZ = 384
    NSDE = 385
    ERFP = 386
    NSDP = 387
    POPI = 388
    ABGT = 389
    ABLT = 390
    SMIA = 391
    SCUR = 392
    SDRV = 393
    REVR = 394
    NSLT = 395
    NSTW = 396
    NSRW = 397
    MNRE = 398
    MXRE = 399
    MNRI = 400
    MXRI = 401
    CEHX = 402
    CEHY = 403
    BLTH = 404
    GBPZ = 405
    NSRD = 406
    NPAF = 407
    MECS = 408
    MECT = 409
    GMTN = 410
    GMTX = 411
    MTFN = 412
    MTFX = 413
    MTHN = 414
    MTHX = 415
    MECA = 416
    MSWN = 417
    MSWX = 418
    GAOI = 419
    MNAI = 420
    MXAI = 421
    HYLD = 422
    OGSS = 423
    SPHD = 424
    SPHS = 425
    SSLP = 426
    SCRV = 427
    DSAG = 428
    DPHS = 429
    DSLP = 430
    DCRV = 431
    EFLA = 432
    CARD = 433
    TSAG = 434
    PSLP = 435
    QSLP = 436



# /// script
# requires-python = ">=3.11"
# dependencies = [
#   "jsonata-python",
#   "pydantic",
#   "zospy @ file://${PROJECT_ROOT}",
# ]
# ///

import json
from math import isnan
from pathlib import Path

from converter import AnalysisDataConverter
from pydantic_core import ValidationError

from zospy import ZOS
from zospy.analyses.base import AnalysisResult

INPUT_FOLDER = Path("tests/data/reference/old")
OUTPUT_FOLDER = Path("tests/data/reference")


def _list_replace_values(data: list, old_values, new_values):
    for old_value, new_value in zip(old_values, new_values):
        if old_value in data:
            data[data.index(old_value)] = new_value


def postprocess_fft_through_focus_mtf(data: dict):
    # Old analyses store enum indices with an offset of 5
    index_to_string = {
        "5": "Modulation",
        "6": "Real",
        "7": "Imaginary",
        "8": "Phase",
        "9": "SquareWave",
    }

    data["settings"]["mtf_type"] = index_to_string[data["settings"]["mtf_type"]]


def postprocess_polarization_pupil_map(data: dict):
    _list_replace_values(data["data"]["pupil_map"]["columns"], ["Phase(Deg)"], ["Phase (Deg)"])


def postprocess_polarization_transmission(data: dict):
    # Remove fields that are not present in the unpolarized case
    if data["settings"]["unpolarized"]:
        data["data"].pop("x_field")
        data["data"].pop("y_field")
        data["data"].pop("x_phase")
        data["data"].pop("y_phase")


def postprocess_single_ray_trace(data: dict):
    for ray_trace_type in ("real_ray_trace_data", "paraxial_ray_trace_data"):
        ray_trace_data = data["data"][ray_trace_type]

        # Rename columns
        _list_replace_values(ray_trace_data["columns"], ("Anglein", "Pathlength"), ("Angle in", "Path length"))

        # Remove last column, which is empty
        ray_trace_data["columns"].pop()
        for row in ray_trace_data["data"]:
            row.pop()
            _list_replace_values(row, ("lensfront", "lensback"), ("lens front", "lens back"))

    # Move the comment for paraxial ray trace data to the last column
    max_row_length = max(len(row) for row in data["data"]["paraxial_ray_trace_data"]["data"])

    # Convert the X-normal column to floats
    if len(data["data"]["real_ray_trace_data"]["columns"]) > 8:
        for row in data["data"]["real_ray_trace_data"]["data"]:
            row[7] = float(row[7].replace(",", ".")) if set(row[7]) != {"-"} else float("nan")

    for row in data["data"]["paraxial_ray_trace_data"]["data"]:
        row.extend([float("nan")] * (max_row_length - len(row)))

        # Get the index of the value before a tail of NaNs
        *_, (last_value_index, last_value) = (
            (i, value) for i, value in enumerate(row) if not (isinstance(value, float) and isnan(value))
        )

        if isinstance(last_value, str) and last_value_index < len(row) - 1:
            row.append(row.pop(last_value_index))


CONVERTERS: list[AnalysisDataConverter] = [
    AnalysisDataConverter(
        old_analysis="geometric_image_analysis",
        new_analysis="GeometricImageAnalysis",
        settings_class="GeometricImageSettings",
        module="zospy.analyses.extendedscene.geometric_image_analysis",
        data_type="dataframe",
    ),
    AnalysisDataConverter(
        old_analysis="fft_through_focus_mtf",
        new_analysis="FFTThroughFocusMTF",
        settings_class="FFTThroughFocusMTFSettings",
        module="zospy.analyses.mtf.fft_through_focus_mtf",
        data_type="dataframe",
        settings_replace_keys={"sample_size": "sampling", "type": "mtf_type"},
        settings_convert_constants={"mtf_type": "Analysis.Settings.Mtf.MtfTypes"},
        postprocess=postprocess_fft_through_focus_mtf,
    ),
    AnalysisDataConverter(
        old_analysis="huygens_mtf",
        new_analysis="HuygensMTF",
        settings_class="HuygensMTFSettings",
        module="zospy.analyses.mtf.huygens_mtf",
        data_type="dataframe",
        settings_replace_keys={"type": "mtf_type"},
    ),
    AnalysisDataConverter(
        old_analysis="physical_optics_propagation",
        new_analysis="PhysicalOpticsPropagation",
        settings_class="PhysicalOpticsPropagationSettings",
        module="zospy.analyses.physicaloptics.physical_optics_propagation",
        data_type="dataframe",
    ),
    AnalysisDataConverter(
        old_analysis="polarization_pupil_map",
        new_analysis="PolarizationPupilMap",
        settings_class="PolarizationPupilMapSettings",
        module="zospy.analyses.polarization.pupil_map",
        data_type="dataclass",
        data_class="PolarizationPupilMapResult",
        data_conversion=r"""
$.data.Data.data{
    "wavelength": Wavelength,
    "field_pos": { "value": FieldPos, "unit": "deg" },
    "x_field": XField,
    "y_field": YField,
    "x_phase": XPhase,
    "y_phase": YPhase,
    "configs": Configs,
    "surface": Surface,
    "transmission": { "value": Transmission, "unit": "%" },
    "pupil_map": Table.data
}""",
        postprocess=postprocess_polarization_pupil_map,
    ),
    AnalysisDataConverter(
        old_analysis="transmission",
        new_analysis="PolarizationTransmission",
        settings_class="PolarizationTransmissionSettings",
        module="zospy.analyses.polarization.transmission",
        data_type="dataclass",
        data_class="PolarizationTransmissionResult",
        data_conversion=r"""
[
    $.data.Settings.data {
        "x_field": Jx,
        "y_field": Jy,
        "x_phase": `X-Phase`,
        "y_phase": `Y-Phase`,
        "grid_size": $replace(Sampling, "x", " x ")
    },
    $.data.Data.data{
        "field_transmissions": [
            {
                "field_pos": {
                    "value": FieldPos,
                    "unit": "deg"
                },
                "total_transmission": TotalTransmission,
                "transmissions": { $string(Wavelength): TotalTransmission }
            }
        ],
        "chief_ray_transmissions": [{
            "field_pos": {
                "value": FieldPos,
                "unit": "deg"
            },
            "wavelength": {
                "1": {
                    "value": Wavelength,
                    "unit": "µm"
                }
            },
            "transmissions": Table.data
        }]
    }
] ~> $merge
""",
        postprocess=postprocess_polarization_transmission,
    ),
    AnalysisDataConverter(
        old_analysis="huygens_psf",
        new_analysis="HuygensPSF",
        settings_class="HuygensPSFSettings",
        module="zospy.analyses.psf.huygens_psf",
        data_type="dataframe",
        settings_replace_keys={
            "pupil_sample_size": "pupil_sampling",
            "image_sample_size": "image_sampling",
            "type": "psf_type",
            "show_as_type": "show_as",
        },
    ),
    AnalysisDataConverter(
        old_analysis="ray_fan",
        new_analysis="RayFan",
        settings_class="RayFanSettings",
        module="zospy.analyses.raysandspots.ray_fan",
        data_type="dataclass",
        data_class="RayFanResult",
        data_conversion=r"""
$map(["Tangential", "Sagittal"], function($d) {
    { $lowercase($d): [$.data.Data.data.$sift(function ($v, $k) { $contains($k, $d) })
        .$each(function($v, $k) {
            {
                "field_number": $number($match($k, /field number (\d+)/).groups[0]),
                "field_coordinate": $match($k, /= (\d+[\.,]\d+) \((\w+)\)/){
                    "value": $number($replace(groups[0], ",", ".")),
                    "unit": groups[1]
                },
                "data": $v.data{
                    "index": data.($[0]),
                    "columns": [$filter(columns, function($v, $i) { $i != 0 }).($replace($, ",", ".").$number())],
                    "data": $map(data, function($v){ $filter($v, function($vv, $i) { $i != 0}) }),
                    "index_names": [columns[0]],
                    "column_names": column_names
                }
            }
        })]
    }
}) ~> $merge
""",
    ),
    AnalysisDataConverter(
        old_analysis="single_ray_trace",
        new_analysis="SingleRayTrace",
        settings_class="SingleRayTraceSettings",
        module="zospy.analyses.raysandspots.single_ray_trace",
        data_type="dataclass",
        data_class="SingleRayTraceResult",
        data_conversion=r"""
[
    {
        "units": "Millimeters",
        "wavelength": $match($.data.RawTextData[$contains($, "Wavelength")], /(\d+[\.,]\d+)\s+([\w\u00B5]+)\s+/){
            "value": groups[0].$replace(",", ".").$number(),
            "unit": groups[1]
        }
    },
    $.data.Settings.data{
        "coordinates": GlobalCoordinates ? "Global coordinates relative to surface 1" : "Local",
        "normalized_x_field_coord": Hx,
        "normalized_y_field_coord": Hy,
        "normalized_x_pupil_coord": Px,
        "normalized_y_pupil_coord": Py
    },
    $.data.Data.data{
        "real_ray_trace_data": RealRayTraceData.data,
        "paraxial_ray_trace_data": ParaxialRayTraceData.data,
        "ym_um_yc_uc_ray_trace_data": null
    }
] ~> $merge
""",
        settings_conversion='$each($.data.Settings.data, function($v, $k) {{ $camelToSnake($k): $v}}) ~> $merge ~> | $ | { "field": $.field = "All" ? 0 } |',
        postprocess=postprocess_single_ray_trace,
    ),
    AnalysisDataConverter(
        old_analysis="cardinal_points",
        new_analysis="CardinalPoints",
        settings_class="CardinalPointsSettings",
        module="zospy.analyses.reports.cardinal_points",
        data_type="dataclass",
        data_class="CardinalPointsResult",
        data_conversion=r"""
(
    $series := $.data.Data.data.$map($zip(index, data), function($v, $i) {{"group": $v[0][0], "property": $v[0][1], "value": $v[1][0]}}){group: {property: value}};
    $series{
        "starting_surface": Surface.`Starting surface`,
        "ending_surface": Surface.`Ending surface`,
        "wavelength": General.Wavelength,
        "orientation": General.Orientation,
        "lens_units": General.`Lens units`,
        "cardinal_points": $merge($each(`Object Space`, function($v, $k){
            {$camelToSnake($k): {
                "object": $v,
                "image": $lookup(`Image Space`, $k)
            }}
        }))
    }
)
""",
        settings_replace_keys={
            "surface1": "surface_1",
            "surface2": "surface_2",
        },
    ),
    AnalysisDataConverter(
        old_analysis="surface_data",
        new_analysis="SurfaceData",
        settings_class="SurfaceDataSettings",
        module="zospy.analyses.reports.surface_data",
        data_type="dataclass",
        data_class="SurfaceDataResult",
        data_conversion=r"""
(
    $series := $.data.Data.data.$map($zip(index, data), function($v, $i) {{"group": $v[0][0], "property": $v[0][1], "value": $v[1][0]}}){group: {$string(property): value}};
    $series.$merge([
        $each(General, function($v, $k) {{ $camelToSnake($k): $v}}) ~> |$|{}, "title"|,
        Surface{"thickness": Thickness, "diameter": Diameter, "edge_thickness": {"x": `X Edge Thick`, "y": `Y Edge Thick`}},
        {
            "material": {
                "indices": [
                    {
                        "number": 1,
                        "wavelength": 0.543,
                        "index": 1.5
                    }
                ],
                "best_fit_glass": IndexOfRefraction.`Best Fit Glass`,
                "glass": null
            },
            "surface_powers": $map($zip(["as_situated", "in_air"], [SurfacePowerAsSituated, SurfacePowerInAir]), function($v){
                {
                    $v[0]: $v[1]{
                        "surf": {"2": `Surf 2`, "3": `Surf 3`},
                        "power": {"2.0,3.0": `Power 2 3`},
                        "efl": {"2.0,3.0": `EFL 2 3`},
                        "f_number": {"2.0,3.0": `F/# 2 3`}
                    }
                }
            }) ~> $merge,
            "shape_factor": Other.`Shape Factor`
        }
    ])
)
""",
    ),
    AnalysisDataConverter(
        old_analysis="system_data",
        new_analysis="SystemData",
        settings_class="SystemDataSettings",
        module="zospy.analyses.reports.system_data",
        data_type="dataclass",
        data_class="SystemDataResult",
        data_conversion=r"""
(
    $replaceKeys := {
        "Effective Focal Length": "effective_focal_length_air",
        "Effective Focal Length 1": "effective_focal_length_image",
        "Primary Wavelength [\u00b5m]": "primary_wavelength",
        "Temperature (C)": "temperature",
        "Pressure (ATM)": "pressure"
    };
    $preprocessKey := function($k){ $k in $replaceKeys.$keys() ? $lookup($replaceKeys, $k) : $camelToSnake($k).$replace("j/e", "j_e").$replace("f/#", "f_number") };
    $preprocessValue := function($v){ $v in ["On", "Off"] ? $v = "On" : $v};
    $.data.Data.data{
        "general_lens_data": GeneralLensData.data.$each(function($v, $k){{ $preprocessKey($k): $preprocessValue($v) }}) ~> $merge
            ~> | $ | $merge([{
                "system_aperture": system_aperture.$split(" = "){ "type": $[0], "value": $[1] },
                "glass_catalogs": glass_catalogs[],
                "apodization": apodization.$match(/(\w+), factor = (\d+[.,]\d+(?:E[+-]\d+))/){ "type": groups[0], "factor": groups[1].$number() },
                "effective_focal_length_air": effective_focal_length_air.$match(/\d+\.\d+/).match.$number(),
                "effective_focal_length_image": effective_focal_length_image.$match(/\d+\.\d+/).match.$number()
            }]), ["include_calculated_data_in_session_file_1"] |,
        "fields": Fields.data{
            "number_of_fields": Info.data.Fields,
            "field_type": Info.data.`Field Type 1`,
            "fields": [$map(Data.data.data, function($v){{"number": $v[0], "x_value": $v[1], "y_value": $v[2], "weight": $v[3] }})]
        },
        "vignetting": [$map(VignettingFactors.data.data, function($v){{"number": $v[0], "vdx": $v[1], "vdy": $v[2], "vcx": $v[3], "vcy": $v[4], "van": $v[5]}})],
        "wavelengths": Wavelengths.data{
            "number_of_wavelengths": Info.data.Wavelengths,
            "units": Info.data.Units,
            "wavelengths": [$map(Data.data.data, function($v){{"number": $v[0], "value": $v[1], "weight": $v[2]}})]
        },
        "abcd_matrix": PredictedCoordinateABCDMatrix.data
})
""",
        settings_conversion="$.data.Settings",
    ),
    AnalysisDataConverter(
        old_analysis="curvature",
        new_analysis="Curvature",
        settings_class="CurvatureSettings",
        module="zospy.analyses.surface.curvature",
        data_type="dataclass",
        data_class="CurvatureResult",
        data_conversion=r"""
{
    "width": 0,
    "decenter_x": 0,
    "decenter_y": 0,
    "decenter_unit": "Millimeters",
    "data": $.data.Data.data
}
""",
        settings_replace_keys={
            "remove_option": "remove",
            "consider_off_axis_aperture": "off_axis_coordinates",
            "best_fit_sphere_options": "bfs_criterion",
            "reverse_direction": "bfs_reverse_direction",
        },
    ),
    AnalysisDataConverter(
        old_analysis="wavefront_map",
        new_analysis="WavefrontMap",
        settings_class="WavefrontMapSettings",
        module="zospy.analyses.wavefront.wavefront_map",
        data_type="dataframe",
        settings_conversion=r'$each($.data.Settings.data, function($v, $k) {{ $camelToSnake($k): $v}}) ~> | $ | {"surface": surface = 0 ? "Image" } | ~> $merge',
    ),
    AnalysisDataConverter(
        old_analysis="zernike_standard_coefficients",
        new_analysis="ZernikeStandardCoefficients",
        settings_class="ZernikeStandardCoefficientsSettings",
        module="zospy.analyses.wavefront.zernike_standard_coefficients",
        data_type="dataclass",
        data_class="ZernikeStandardCoefficientsResult",
        data_conversion=r"""
[
    $.data.Settings.data{
        "subaperture_decenter_sx": Sx,
        "subaperture_decenter_sy": Sy,
        "subaperture_radius_sr": Sr
    },
    $.data.Data.data.GeneralData.data.$map($zip(index, data), function($v, $i) {{$camelToSnake($v[0]):  {"value": $v[1][0], "unit": $v[1][1]}}}){
        "surface": surface.value,
        "field": field{"value": value, "unit": unit.$replace(/[\(\)]/, "")},
        "wavelength": wavelength,
        "peak_to_valley_to_chief": peak_to_valley_to_chief,
        "peak_to_valley_to_centroid": peak_to_valley_to_centroid,
        "from_integration_of_the_fitted_coefficients": $merge([
            $map(["rms_to_chief", "rms_to_centroid", "variance"], function($v){ {$v: $lookup($, $v)} }),
            {"strehl_ratio": $.strehl_ratio_est.value }
        ]),
        "from_integration_of_the_rays": $merge([
            $map(["rms_to_chief", "rms_to_centroid", "variance"], function($v){ {$v: $lookup($, $v)} }),
            {"strehl_ratio": $.strehl_ratio_est.value }
        ]),
        "rms_fit_error": rms_fit_error,
        "maximum_fit_error": maximum_fit_error
    },
    $.data.Data.data.Coefficients.data{
        "coefficients": $map(data, function($v, $i){ {($i + 1).$string(): {"value": $v[0], "formula": $v[2] }}}) ~>
        $merge
    }
] ~> $merge
""",
        settings_replace_keys={
            "sample_size": "sampling",
            "maximum_number_of_terms": "maximum_terms",
        },
    ),
]


def convert_reference_data_files(converter: AnalysisDataConverter):
    for file in INPUT_FOLDER.glob(f"*{converter.old_analysis}*.json"):
        reference_data = json.loads(file.read_text())

        converted_data = converter.convert_data(reference_data)

        # Attempt to load the data as an AnalysisResult to ensure it is valid
        try:
            AnalysisResult.from_json(json.dumps(converted_data, indent=4))
        except ValidationError as e:
            print(f"Error converting {file.name}: {e}")

        output_file = OUTPUT_FOLDER / file.name
        output_file.write_text(json.dumps(converted_data, indent=4))


if __name__ == "__main__":
    # Initialize ZOS instance, to check constants existence
    zos = ZOS()

    for converter in CONVERTERS:
        convert_reference_data_files(converter)



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

__all__ = "DataType"

class DataType:
    Wavefront = 0
    SpotRadius = 1
    Spot_X = 2
    Spot_Y = 3
    StrehlRation = 4



from typing import Any

import pytest
from lark import Lark
from lark.load_grammar import FromPackageLoader
from lark.visitors import merge_transformers

from zospy.analyses.parsers.transformers import ParametricField, SimpleField, UnitField, ZospyTransformer
from zospy.api import config

zospy_grammar_loader = FromPackageLoader("zospy", ["analyses/parsers/grammars"])


class SingleTokenZospyTransformer(ZospyTransformer):
    def start(self, args):
        return args[0]


transformer = merge_transformers(SingleTokenZospyTransformer(), zospy=ZospyTransformer())


def load_token(token: str) -> Lark:
    grammar = f"""start: {token}

%import common.WS_INLINE
%import zospy.{token}
%ignore WS_INLINE
"""
    return Lark(grammar, parser="earley", start="start", import_paths=[zospy_grammar_loader])


def parse_token(token: str, text: str) -> Any:
    parser = load_token(token)

    return transformer.transform(parser.parse(text))


@pytest.mark.parametrize(
    "value",
    ["01-01-2021", "01/01/2024", "05.12.1234"],
)
def test_date(value):
    # Dates are returned as is
    assert parse_token("DATE", value) == value


@pytest.mark.parametrize(
    "value,expected,decimal_point",
    [
        # Point as decimal separator
        ("1.23", 1.23, "."),
        (".23", 0.23, "."),
        ("+1.234", 1.234, "."),
        ("-1.234", -1.234, "."),
        ("1.234e-2", 1.234e-2, "."),
        ("1.234e+2", 1.234e2, "."),
        ("1.234E-2", 1.234e-2, "."),
        ("1.234E+2", 1.234e2, "."),
        ("-1.234e-2", -1.234e-2, "."),
        ("-1.234e+2", -1.234e2, "."),
        ("-1.234E-2", -1.234e-2, "."),
        ("-1.234E+2", -1.234e2, "."),
        # Comma as decimal separator
        ("1,23", 1.23, ","),
        (",23", 0.23, ","),
        ("+1,234", 1.234, ","),
        ("-1,234", -1.234, ","),
        ("1,234e-2", 1.234e-2, ","),
        ("1,234e+2", 1.234e2, ","),
        ("1,234E-2", 1.234e-2, ","),
        ("1,234E+2", 1.234e2, ","),
        ("-1,234e-2", -1.234e-2, ","),
        ("-1,234e+2", -1.234e2, ","),
        ("-1,234E-2", -1.234e-2, ","),
        ("-1,234E+2", -1.234e2, ","),
        ("infinity", float("inf"), "."),
    ],
)
def test_float(value, expected, decimal_point, monkeypatch):
    monkeypatch.setattr(config, "DECIMAL_POINT", decimal_point)
    monkeypatch.setattr(config, "THOUSANDS_SEPARATOR", "")

    result = parse_token("FLOAT", value)
    assert isinstance(result, float)
    assert result == expected


@pytest.mark.parametrize(
    "value,expected",
    [
        ("1", 1),
        ("+1", 1),
        ("-1", -1),
        ("1.23", 1.23),
        ("+1.23", 1.23),
        ("-1.23", -1.23),
        ("1.23e-2", 1.23e-2),
        ("1.23e+2", 1.23e2),
        ("1.23E-2", 1.23e-2),
        ("1.23E+2", 1.23e2),
        ("-1.23e-2", -1.23e-2),
        ("-1.23e+2", -1.23e2),
        ("-1.23E-2", -1.23e-2),
        ("-1.23E+2", -1.23e2),
    ],
)
def test_number(value, expected, monkeypatch):
    monkeypatch.setattr(config, "DECIMAL_POINT", ".")
    monkeypatch.setattr(config, "THOUSANDS_SEPARATOR", "")

    result = parse_token("_number", value)

    assert result == expected
    assert type(result) is type(expected)


@pytest.mark.parametrize(
    "value,expected",
    [
        ("Alice Bob Charlie", "Alice Bob Charlie"),
        ("Multiple  spaces", "Multiple spaces"),
        (" Leading space", "Leading space"),
        ("Trailing space ", "Trailing space"),
        ("  Leading and  trailing  spaces  ", "Leading and trailing spaces"),
        ("Tabs\tand\ttabs", "Tabs and tabs"),
    ],
)
def test_multi_string(value, expected):
    result = parse_token("multi_string", value)
    assert result == expected


def test_string_list():
    grammar = """start: value

!value: "abc" "def" /Test with space/ "ghi" -> string_list

%import common.WS_INLINE
%ignore WS_INLINE
"""

    parser = Lark(grammar, parser="earley", start="start")
    result = transformer.transform(parser.parse("abc def Test with space ghi"))

    assert result == ["abc", "def", "Test with space", "ghi"]


def test_list():
    grammar = """start: value

    value: INT+ -> list

    %import common.WS_INLINE
    %import zospy.INT
    %ignore WS_INLINE
    """

    parser = Lark(grammar, parser="earley", start="start", import_paths=[zospy_grammar_loader])
    result = transformer.transform(parser.parse("1 2 3 4 5"))

    assert result == [1, 2, 3, 4, 5]


@pytest.fixture
def setup_field_test(monkeypatch):
    monkeypatch.setattr(config, "DECIMAL_POINT", ".")
    monkeypatch.setattr(config, "THOUSANDS_SEPARATOR", "")

    def setup(value):
        if not value.endswith("\n"):
            value += "\n"

        return value

    return setup


@pytest.mark.parametrize(
    "value,expected",
    [
        ("Name: value", SimpleField("Name", "value")),
        ("Name with spaces: 123", SimpleField("Name with spaces", 123)),
        ("Name_with_underscores: 1.23", SimpleField("Name_with_underscores", 1.23)),
        ("Name-with-dashes: multi string", SimpleField("Name-with-dashes", "multi string")),
        ("Name (with parentheses): -1.23E-5", SimpleField("Name (with parentheses)", -1.23e-5)),
        ("Space before colon : 1.23", SimpleField("Space before colon", 1.23)),
        ("Tabs around colon\t:\t01-01-1970", SimpleField("Tabs around colon", "01-01-1970")),
        ("Dot-separated date: 01.01.1970", SimpleField("Dot-separated date", "01.01.1970")),
    ],
)
def test_simple_field(value, expected, setup_field_test):
    value = setup_field_test(value)

    result = parse_token("simple_field", value)

    assert result == expected
    assert isinstance(result.value, type(expected.value))


@pytest.mark.parametrize(
    "value, expected",
    [
        ("Name 123: 1.23", SimpleField("Name", ParametricField(123, 1.23))),
        ("Double parameters 123 4.56: 7.89", SimpleField("Double parameters", ParametricField((123, 4.56), 7.89))),
        (
            "Negative parameters -3.14 -2: a meaningless string",
            SimpleField("Negative parameters", ParametricField((-3.14, -2), "a meaningless string")),
        ),
    ],
)
def test_parametric_field(value, expected, setup_field_test):
    value = setup_field_test(value)

    result: SimpleField[ParametricField] = parse_token("parametric_field", value)

    assert result == expected
    assert isinstance(result.value.value, type(expected.value.value))

    if isinstance(result.value.parameters, tuple):
        for r, e in zip(result.value.parameters, expected.value.parameters):
            assert isinstance(r, type(e))
    else:
        assert isinstance(result.value.parameters, type(expected.value.parameters))


@pytest.mark.parametrize(
    "value, expected",
    [
        ("Length: 10m", SimpleField("Length", UnitField(value=10, unit="m"))),
        ("Weight: 5 kg", SimpleField("Weight", UnitField(value=5, unit="kg"))),
        (
            "Temperature: 100 space separated unit",
            SimpleField("Temperature", UnitField(value=100, unit="space separated unit")),
        ),
        ("Wavelength: 0.543 µm", SimpleField("Wavelength", UnitField(value=0.543, unit="µm"))),
        ("Something: 30 (parenthesized)", SimpleField("Something", UnitField(value=30, unit="parenthesized"))),
    ],
)
def test_unit_field(value, expected, setup_field_test):
    value = setup_field_test(value)

    result = parse_token("unit_field", value)

    assert result == expected
    assert isinstance(result.value, type(expected.value))


@pytest.mark.parametrize(
    "value, expected",
    [
        (
            "Transmission 123: 456 mm",
            SimpleField("Transmission", ParametricField(123, UnitField(value=456, unit="mm"))),
        ),
        ("Length 1 2: 3 m", SimpleField("Length", ParametricField((1, 2), UnitField(value=3, unit="m")))),
    ],
)
def test_parametric_unit_field(value, expected, setup_field_test):
    value = setup_field_test(value)

    result = parse_token("parametric_unit_field", value)

    assert result == expected
    assert isinstance(result.value.value, type(expected.value.value))

    if isinstance(result.value.parameters, tuple):
        for r, e in zip(result.value.parameters, expected.value.parameters):
            assert isinstance(r, type(e))
    else:
        assert isinstance(result.value.parameters, type(expected.value.parameters))


@pytest.mark.parametrize(
    "value, expected",
    [
        ("Simple field : 1.23", SimpleField("Simple field", 1.23)),
        ("Parametric field 1 2: 3.14", SimpleField("Parametric field", ParametricField((1, 2), 3.14))),
        ("Unit field: 10 m", SimpleField("Unit field", UnitField(value=10, unit="m"))),
        (
            "Parametric unit field 1 2: 3 m",
            SimpleField("Parametric unit field", ParametricField((1, 2), UnitField(value=3, unit="m"))),
        ),
    ],
)
def test_field(value, expected, setup_field_test):
    value = setup_field_test(value)

    result = parse_token("_field", value)

    assert result == expected
    assert isinstance(result.value, type(expected.value))


FIELD_GROUP_VALUE = """Group key:
    Simple field : 1.23
    Parametric field 1 2: 3.14
    Parametric unit field 3.14 : 1.23 m
"""

FIELD_GROUP_GRAMMAR = r"""start: _fg

_fg: field_group{name, fields}
!name: "Group" "key" -> field_name
fields: _field+ -> dict

%import zospy (_field, field_group)
%import common.WS_INLINE
%ignore WS_INLINE
"""


def test_field_group(setup_field_test):  # noqa: ARG001
    parser = Lark(FIELD_GROUP_GRAMMAR, parser="earley", start="start", import_paths=[zospy_grammar_loader])

    result = transformer.transform(parser.parse(FIELD_GROUP_VALUE))

    assert result == SimpleField(
        "Group key",
        {
            "Simple field": 1.23,
            "Parametric field": {(1, 2): 3.14},
            "Parametric unit field": {3.14: {"value": 1.23, "unit": "m"}},
        },
    )


TABLE_VALUE = """Alice Bob Charlie
1 2.3 4 5.6
7 8.9 10 11.12
"""

TABLE_GRAMMAR = r"""start: _table

_table: table{header, row}
!header: "Alice" "Bob" "Charlie" -> string_list
row: _numbers -> list

%import zospy (_field, field_group, _numbers, table)
%import common.NEWLINE -> _NEWLINE
%import common.WS_INLINE
%ignore WS_INLINE
"""


def test_table(setup_field_test):  # noqa: ARG001
    parser = Lark(TABLE_GRAMMAR, parser="earley", start="start", import_paths=[zospy_grammar_loader])

    result = transformer.transform(parser.parse(TABLE_VALUE))

    assert result == (["Alice", "Bob", "Charlie"], [[1, 2.3, 4, 5.6], [7, 8.9, 10, 11.12]])


@pytest.mark.parametrize(
    "value, expected, tokens, expected_token",
    [
        ("1.23", 1.23, ["STRING", "FLOAT"], "FLOAT"),
        ("01-02-2023", "01-02-2023", ["STRING", "DATE"], "DATE"),
        ("01.02.2923", "01.02.2923", ["STRING", "DATE", "FLOAT"], "DATE"),
        ("Simple field : 123\n", SimpleField("Simple field", 123), ["STRING", "simple_field"], "simple_field"),
    ],
)
def test_token_precedence(value, expected, tokens, expected_token, setup_field_test):  # noqa: ARG001
    grammar = f"""start: _value _NEWLINE?

_value: {" | ".join(tokens)}

%import common.NEWLINE -> _NEWLINE
%import common.WS_INLINE
%import zospy ({", ".join(tokens)})
%ignore WS_INLINE
"""

    parser = Lark(grammar, parser="earley", start="start", import_paths=[zospy_grammar_loader])
    parse_result = parser.parse(value)
    result = transformer.transform(parse_result)

    assert result == expected
    if expected_token.isupper():
        # Token is a terminal
        assert parse_result.children[0].type == expected_token
    else:
        # Token is a rule
        assert str(parse_result.children[0].data) == expected_token



"""Script to run an examples and save the output.

This script runs the specified example using the `run_single_example.py` script and saves the output.
`uv` is used to run the example in an isolated environment, with only the required dependencies.
"""

from __future__ import annotations

import argparse
import sys
from pathlib import Path

from _common import process_example


def main(args: argparse.Namespace) -> int:
    example_directory: Path = args.example_directory.resolve()
    zospy_location = args.zospy_location.resolve(strict=False) if args.zospy_location is not None else None
    output_folder = args.output.resolve(strict=False) if args.output is not None else example_directory
    keep_extension_mode = args.keep_extension_mode

    if not example_directory.exists():
        print(
            f"Directory {example_directory} does not exist. Note: when run using Hatch, paths with backward slashes "
            "may not be found. Replace backward slashes with forward slashes in the path."
        )
        return 1

    if not example_directory.is_dir():
        print(f"{example_directory} is not a directory. Please provide a directory.")
        return 1

    success = process_example(
        example_directory,
        output_folder,
        editable_zospy_location=zospy_location,
        keep_extension_mode=keep_extension_mode,
    )

    if success:
        print(f"\n\nExample {example_directory} ran successfully.")
        return 0

    print(f"\n\nExample {example_directory} failed to run.")
    return 1


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Run an example and save the output.")
    (
        parser.add_argument(
            "example_directory",
            type=Path,
            help="Path to the directory containing the example.",
        ),
    )
    parser.add_argument(
        "--zospy-location",
        type=Path,
        help="Path to the directory containing ZOSPy. If not provided, ZOSPy is installed from PyPI.",
        default=None,
    )
    parser.add_argument(
        "--output",
        type=Path,
        help="Path to save the output to. Defaults to the input example directory.",
        default=None,
    )
    parser.add_argument(
        "--keep-extension-mode",
        action="store_true",
        help="Do not change extension mode to standalone mode prior to running the notebook.",
    )

    args = parser.parse_args()

    sys.exit(main(args))



"""3D system viewer."""

from __future__ import annotations

from typing import TYPE_CHECKING, Annotated, Literal

from pydantic import Field

from zospy.analyses.decorators import analysis_settings
from zospy.analyses.parsers.types import FieldNumber, WavelengthNumber, ZOSAPIConstant  # noqa: TCH001
from zospy.analyses.systemviewers.base import ImageSize, SystemViewerWrapper
from zospy.api import constants

if TYPE_CHECKING:
    from zospy.api import _ZOSAPI

__all__ = ("Viewer3D", "Viewer3DSettings")


@analysis_settings
class Viewer3DSettings:
    """Settings for the 3D system viewer.

    Attributes
    ----------
    start_surface : int, optional
        The starting surface index for the 3D viewer. Defaults to 1.
    end_surface : int, optional
        The ending surface index for the 3D viewer. A value of -1 indicates the last surface. Defaults to -1.
    number_of_rays : int, optional
        The number of rays to be used in the analysis. Defaults to 3.
    wavelength : int | str, optional
        The wavelength index to be used. Can be an integer or "All" for all wavelengths. Defaults to "All".
    field : int | str, optional
        The field index to be used. Can be an integer or "All" for all fields. Defaults to "All".
    ray_pattern : constants.Tools.General.RayPatternType | str, optional
        The ray pattern to be used in the analysis. Defaults to "XYFan".
    color_rays_by : constants.Tools.Layouts.ColorRaysByOptions | str, optional
        The criterion for coloring rays in the analysis. Defaults to "Fields".
    delete_vignetted : bool, optional
        Flag indicating whether to delete vignetted rays. Defaults to False.
    hide_lens_faces : bool, optional
        Flag indicating whether to hide lens faces. Defaults to False.
    hide_lens_edges : bool, optional
        Flag indicating whether to hide lens edges. Defaults to False.
    hide_x_bars : bool, optional
        Flag indicating whether to hide the x-component of lens faces. Defaults to False.
    draw_paraxial_pupils : bool, optional
        Flag indicating whether to draw paraxial entrance and exit pupils. Defaults to False.
    fletch_rays : bool, optional
        Flag indicating whether to draw small arrows indicating the direction of the rays. Defaults to False.
    split_nsc_rays : bool, optional
        Flag indicating whether to split rays from non-sequential sources at ray-surface intercepts. Defaults to False.
    scatter_nsc_rays : bool, optional
        Flag indicating whether to scatter rays from non-sequential sources at ray-surface intercepts.
        Defaults to False.
    draw_real_entrance_pupils : constants.Tools.Layouts.RealPupilOptions | str, optional
        How to draw real entrance pupils. Defaults to "Pupils_Off". Can be one of ['Pupils_Off', 'Pupils_4',
        'Pupils_8', 'Pupils_16', 'Pupils_32'].
    draw_real_exit_pupils : constants.Tools.Layouts.RealPupilOptions | str, optional
        How to draw real exit pupils. Defaults to "Pupils_Off". Can be one of ['Pupils_Off', 'Pupils_4', 'Pupils_8',
        'Pupils_16', 'Pupils_32'].
    surface_line_thickness : constants.Tools.Layouts.LineThicknessOptions | str, optional
        The thickness of the lines for the surfaces. Defaults to "Standard".
    rays_line_thickness : constants.Tools.Layouts.LineThicknessOptions | str, optional
        The thickness of the lines for the rays. Defaults to "Standard".
    configuration_all : bool, optional
        Flag indicating whether to use all configurations, if multiple configurations are present. Defaults to False.
        If multiple configurations are displayed, the `configuration_offset_*` parameters can be used to add offsets
        between the different configurations.
    configuration_current : bool, optional
        Flag indicating whether to only display the current configuration, when multiple configurations are present.
        Defaults to False.
    configuration_offset_x : float, optional
        The offset along the X-axis between configurations, if multiple configurations are present. Defaults to 0.
    configuration_offset_y : float, optional
        The offset along the Y-axis between configurations, if multiple configurations are present. Defaults to 0.
    configuration_offset_z : float, optional
        The offset along the Z-axis between configurations, if multiple configurations are present. Defaults to 0.
    camera_viewpoint_angle_x : float, optional
        Rotation of the system around the X-axis, in degrees. Defaults to 0.
    camera_viewpoint_angle_y : float, optional
        Rotation of the system around the Y-axis, in degrees. Defaults to 0.
    camera_viewpoint_angle_z : float, optional
        Rotation of the system around the Z-axis, in degrees. Defaults to 0.
    image_size : tuple[int, int], optional
        The size of the output image in pixels (width, height). Defaults to (800, 600).
    """

    start_surface: int = Field(default=1, ge=1, description="Starting surface number")
    end_surface: Literal[-1] | Annotated[int, Field(ge=1)] = Field(default=-1, description="Ending surface number")
    number_of_rays: int = Field(default=3, ge=1, description="Number of rays")
    wavelength: WavelengthNumber = Field(default="All", description="Wavelength number")
    field: FieldNumber = Field(default="All", description="Field number")
    # ray_pattern: str = Field(default="XYFan", description="Ray pattern")
    ray_pattern: ZOSAPIConstant("Tools.General.RayPatternType") = Field(default="XYFan", description="Ray pattern")
    # color_rays_by: str = Field(default="Fields", description="Color rays by")
    color_rays_by: ZOSAPIConstant("Tools.Layouts.ColorRaysByOptions") = Field(
        default="Fields", description="Color rays by"
    )
    delete_vignetted: bool = Field(default=False, description="Delete vignetted rays")
    hide_lens_faces: bool = Field(default=False, description="Hide lens faces")
    hide_lens_edges: bool = Field(default=False, description="Hide lens edges")
    hide_x_bars: bool = Field(default=False, description="Hide X bars")
    draw_paraxial_pupils: bool = Field(default=False, description="Draw paraxial pupils")
    fletch_rays: bool = Field(default=False, description="Fletch rays")
    split_nsc_rays: bool = Field(default=False, description="Split NSC rays")
    scatter_nsc_rays: bool = Field(default=False, description="Scatter NSC rays")
    # draw_real_entrance_pupils: str = Field(default="Pupils_Off", description="Draw real entrance pupils")
    draw_real_entrance_pupils: ZOSAPIConstant("Tools.Layouts.RealPupilOptions") = Field(
        default="Pupils_Off", description="Draw real entrance pupils"
    )
    # draw_real_exit_pupils: str = Field(default="Pupils_Off", description="Draw real exit pupils")
    draw_real_exit_pupils: ZOSAPIConstant("Tools.Layouts.RealPupilOptions") = Field(
        default="Pupils_Off", description="Draw real exit pupils"
    )
    # surface_line_thickness: str = Field(default="Standard", description="Surface line thickness")
    surface_line_thickness: ZOSAPIConstant("Tools.Layouts.LineThicknessOptions") = Field(
        default="Standard", description="Surface line thickness"
    )
    # rays_line_thickness: str = Field(default="Standard", description="Rays line thickness")
    rays_line_thickness: ZOSAPIConstant("Tools.Layouts.LineThicknessOptions") = Field(
        default="Standard", description="Rays line thickness"
    )
    configuration_all: bool = Field(default=False, description="Draw all configurations")
    configuration_current: bool = Field(default=False, description="Draw only current configuration")
    configuration_offset_x: float = Field(default=0, description="Configuration X offset")
    configuration_offset_y: float = Field(default=0, description="Configuration Y offset")
    configuration_offset_z: float = Field(default=0, description="Configuration Z offset")
    camera_viewpoint_angle_x: float = Field(default=0, description="Camera viewpoint X angle")
    camera_viewpoint_angle_y: float = Field(default=0, description="Camera viewpoint Y angle")
    camera_viewpoint_angle_z: float = Field(default=0, description="Camera viewpoint Z angle")
    image_size: ImageSize = Field(default=(800, 600), description="Image size")


class Viewer3D(SystemViewerWrapper[Viewer3DSettings], analysis_type="Draw3D", mode="Sequential"):
    """3D system viewer."""

    def __init__(
        self,
        *,
        start_surface: int = 1,
        end_surface: int = -1,
        number_of_rays: int = 3,
        wavelength: int | Literal["All"] = "All",
        field: int | Literal["All"] = "All",
        ray_pattern: constants.Tools.General.RayPatternType | str = "XYFan",
        color_rays_by: constants.Tools.Layouts.ColorRaysByOptions | str = "Fields",
        delete_vignetted: bool = False,
        hide_lens_faces: bool = False,
        hide_lens_edges: bool = False,
        hide_x_bars: bool = False,
        draw_paraxial_pupils: bool = False,
        fletch_rays: bool = False,
        split_nsc_rays: bool = False,
        scatter_nsc_rays: bool = False,
        draw_real_entrance_pupils: constants.Tools.Layouts.RealPupilOptions | str = "Pupils_Off",
        draw_real_exit_pupils: constants.Tools.Layouts.RealPupilOptions | str = "Pupils_Off",
        surface_line_thickness: constants.Tools.Layouts.LineThicknessOptions | str = "Standard",
        rays_line_thickness: constants.Tools.Layouts.LineThicknessOptions | str = "Standard",
        configuration_all: bool = False,
        configuration_current: bool = False,
        configuration_offset_x: float = 0,
        configuration_offset_y: float = 0,
        configuration_offset_z: float = 0,
        camera_viewpoint_angle_x: float = 0,
        camera_viewpoint_angle_y: float = 0,
        camera_viewpoint_angle_z: float = 0,
        image_size: tuple[int, int] = (800, 600),
    ):
        """Create a new 3D system viewer.

        See Also
        --------
        Viewer3DSettings : Settings for the 3D system viewer.
        """
        super().__init__(settings_kws=locals())

    def configure_layout_tool(
        self,
    ) -> _ZOSAPI.Tools.Layouts.I3DViewerExport:
        """Configure the  3D viewer."""
        layout_tool = self.oss.Tools.Layouts.Open3DViewerExport()

        layout_tool.StartSurface = self.settings.start_surface
        layout_tool.EndSurface = self._validate_end_surface(self.settings.start_surface, self.settings.end_surface)
        layout_tool.NumberOfRays = self.settings.number_of_rays
        layout_tool.Wavelength = self._validate_wavelength(self.settings.wavelength)
        layout_tool.Field = self._validate_field(self.settings.field)
        layout_tool.RayPattern = constants.process_constant(
            constants.Tools.General.RayPatternType, self.settings.ray_pattern
        )
        layout_tool.ColorRaysBy = constants.process_constant(
            constants.Tools.Layouts.ColorRaysByOptions, self.settings.color_rays_by
        )
        layout_tool.DeleteVignetted = self.settings.delete_vignetted
        layout_tool.HideLensFaces = self.settings.hide_lens_faces
        layout_tool.HideLensEdges = self.settings.hide_lens_edges
        layout_tool.HideXBars = self.settings.hide_x_bars
        layout_tool.DrawParaxialPupils = self.settings.draw_paraxial_pupils
        layout_tool.FletchRays = self.settings.fletch_rays
        layout_tool.SplitNSCRays = self.settings.split_nsc_rays
        layout_tool.ScatterNSCRays = self.settings.scatter_nsc_rays
        layout_tool.DrawRealEntrancePupils = constants.process_constant(
            constants.Tools.Layouts.RealPupilOptions, self.settings.draw_real_entrance_pupils
        )
        layout_tool.DrawRealExitPupils = constants.process_constant(
            constants.Tools.Layouts.RealPupilOptions, self.settings.draw_real_exit_pupils
        )
        layout_tool.SurfaceLineThickness = constants.process_constant(
            constants.Tools.Layouts.LineThicknessOptions, self.settings.surface_line_thickness
        )
        layout_tool.RaysLineThickness = constants.process_constant(
            constants.Tools.Layouts.LineThicknessOptions, self.settings.rays_line_thickness
        )
        layout_tool.ConfigurationAll = self.settings.configuration_all
        layout_tool.ConfigurationCurrent = self.settings.configuration_current
        layout_tool.ConfigurationOffsetX = self.settings.configuration_offset_x
        layout_tool.ConfigurationOffsetY = self.settings.configuration_offset_y
        layout_tool.ConfigurationOffsetZ = self.settings.configuration_offset_z
        layout_tool.CameraViewpointAngleX = self.settings.camera_viewpoint_angle_x
        layout_tool.CameraViewpointAngleY = self.settings.camera_viewpoint_angle_y
        layout_tool.CameraViewpointAngleZ = self.settings.camera_viewpoint_angle_z
        layout_tool.OutputPixelWidth, layout_tool.OutputPixelHeight = self.settings.image_size

        return layout_tool



"""Cross section (2D) viewer."""

from __future__ import annotations

from typing import TYPE_CHECKING, Annotated, Literal

from pydantic.dataclasses import Field

from zospy.analyses.decorators import analysis_settings
from zospy.analyses.parsers.types import FieldNumber, WavelengthNumber, ZOSAPIConstant  # noqa: TCH001
from zospy.analyses.systemviewers.base import ImageSize, SystemViewerWrapper
from zospy.api import constants

if TYPE_CHECKING:
    from zospy.api import _ZOSAPI

__all__ = ("CrossSection", "CrossSectionSettings")


@analysis_settings
class CrossSectionSettings:
    """Settings for the cross section viewer.

    Attributes
    ----------
    start_surface : int, optional
        The starting surface index for the cross-section analysis. Defaults to 1.
    end_surface : int, optional
        The ending surface index for the cross-section analysis. A value of -1 indicates the last surface.
        Defaults to -1.
    number_of_rays : int, optional
        The number of rays to be used in the analysis. Defaults to 3.
    y_stretch : float, optional
        The stretch factor in the Y-axis for the analysis visualization. Defaults to 1.0.
    fletch_rays : bool, optional
        Flag indicating whether to fletch rays. Defaults to False.
    wavelength : int | str, optional
        The wavelength index to be used. Can be an integer or "all" for all wavelengths. Defaults to "all".
    field : int | str, optional
        The field index to be used. Can be an integer or "all" for all fields. Defaults to "all".
    color_rays_by : constants.Tools.Layouts.ColorRaysByCrossSectionOptions | str, optional
        The criterion for coloring rays in the analysis. Defaults to "Fields".
    upper_pupil : float, optional
        The upper pupil limit for the analysis. Defaults to 1.
    lower_pupil : float, optional
        The lower pupil limit for the analysis. Defaults to -1.
    delete_vignetted : bool, optional
        Flag indicating whether to delete vignetted rays. Defaults to False.
    marginal_and_chief_only : bool, optional
        Flag indicating whether to include only marginal and chief rays in the analysis. Defaults to False.
    image_size : tuple[int, int], optional
        The size of the output image in pixels (width, height). Defaults to (800, 600).
    rays_line_thickness : constants.Tools.Layouts.LineThicknessOptions | str, optional
        The thickness of the lines in the output visualization. Defaults to "Standard".
    surface_line_thickness : constants.Tools.Layouts.LineThicknessOptions | str, optional
        The thickness of the surface lines in the output visualization. Defaults to "Standard".
    """

    start_surface: int = Field(default=1, ge=1, description="Starting surface number")
    end_surface: Literal[-1] | Annotated[int, Field(ge=1)] = Field(default=-1, description="Ending surface number")
    number_of_rays: int = Field(default=3, ge=1, description="Number of rays")
    y_stretch: float = Field(default=1.0, ge=0, description="Y stretch factor")
    fletch_rays: bool = Field(default=False, description="Fletch rays")
    wavelength: WavelengthNumber = Field(default="All", description="Wavelength number")
    field: FieldNumber = Field(default="All", description="Field number")
    color_rays_by: ZOSAPIConstant("Tools.Layouts.ColorRaysByCrossSectionOptions") = Field(
        default="Fields", description="Color rays by"
    )
    upper_pupil: float = Field(default=1.0, gt=-1, le=1, description="Upper pupil limit")
    lower_pupil: float = Field(default=-1.0, ge=-1, lt=1, description="Lower pupil limit")
    delete_vignetted: bool = Field(default=False, description="Delete vignetted rays")
    marginal_and_chief_only: bool = Field(default=False, description="Draw marginal and chief rays only")
    image_size: ImageSize = Field(default=(800, 600), description="Image size")
    rays_line_thickness: ZOSAPIConstant("Tools.Layouts.LineThicknessOptions") = Field(
        default="Standard", description="Rays line thickness"
    )
    surface_line_thickness: ZOSAPIConstant("Tools.Layouts.LineThicknessOptions") = Field(
        default="Standard", description="Surface line thickness"
    )


class CrossSection(SystemViewerWrapper[CrossSectionSettings], analysis_type="Draw2D", mode="Sequential"):
    """Cross section viewer."""

    def __init__(
        self,
        *,
        start_surface: int = 1,
        end_surface: int = -1,
        number_of_rays: int = 3,
        y_stretch: float = 1.0,
        fletch_rays: bool = False,
        wavelength: int | Literal["All"] = "All",
        field: int | Literal["All"] = "All",
        color_rays_by: constants.Tools.Layouts.ColorRaysByCrossSectionOptions | str = "Fields",
        upper_pupil: float = 1.0,
        lower_pupil: float = -1.0,
        delete_vignetted: bool = False,
        marginal_and_chief_only: bool = False,
        image_size: tuple[int, int] = (800, 600),
        rays_line_thickness: constants.Tools.Layouts.LineThicknessOptions | str = "Standard",
        surface_line_thickness: constants.Tools.Layouts.LineThicknessOptions | str = "Standard",
    ):
        """Create a new cross section viewer.

        See Also
        --------
        CrossSectionSettings : Settings for the cross section viewer.
        """
        super().__init__(settings_kws=locals())

    def configure_layout_tool(self) -> _ZOSAPI.Tools.Layouts.ICrossSectionExport:
        """Configure the cross section viewer."""
        layout_tool = self.oss.Tools.Layouts.OpenCrossSectionExport()
        layout_tool.StartSurface = self.settings.start_surface
        layout_tool.EndSurface = self._validate_end_surface(self.settings.start_surface, self.settings.end_surface)
        layout_tool.NumberOfRays = self.settings.number_of_rays
        layout_tool.YStretch = self.settings.y_stretch
        layout_tool.FletchRays = self.settings.fletch_rays
        layout_tool.Wavelength = self._validate_wavelength(self.settings.wavelength)
        layout_tool.Field = self._validate_field(self.settings.field)
        layout_tool.ColorRaysBy = constants.process_constant(
            constants.Tools.Layouts.ColorRaysByCrossSectionOptions, self.settings.color_rays_by
        )
        layout_tool.UpperPupil = self.settings.upper_pupil
        layout_tool.LowerPupil = self.settings.lower_pupil
        layout_tool.DeleteVignetted = self.settings.delete_vignetted
        layout_tool.MarginalAndChiefOnly = self.settings.marginal_and_chief_only
        layout_tool.OutputPixelHeight, layout_tool.OutputPixelWidth = self.settings.image_size
        layout_tool.RaysLineThickness = constants.process_constant(
            constants.Tools.Layouts.LineThicknessOptions, self.settings.rays_line_thickness
        )
        layout_tool.SurfaceLineThickness = constants.process_constant(
            constants.Tools.Layouts.LineThicknessOptions, self.settings.surface_line_thickness
        )

        return layout_tool



"""Surface Curvature analysis."""

from __future__ import annotations

import logging
import re
from typing import Annotated

from pydantic import Field

from zospy.analyses.base import BaseAnalysisWrapper
from zospy.analyses.decorators import analysis_result, analysis_settings
from zospy.analyses.parsers.types import ValidatedDataFrame, ZOSAPIConstant  # noqa: TCH001
from zospy.api import config, constants
from zospy.utils.pyutils import atox
from zospy.utils.zputils import standardize_sampling, unpack_datagrid

__all__ = ("Curvature", "CurvatureSettings")


@analysis_result
class CurvatureResult:
    width: float
    decenter_x: float
    decenter_y: float
    decenter_unit: str
    data: ValidatedDataFrame


@analysis_settings
class CurvatureSettings:
    """Settings for the Curvature analysis.

    Attributes
    ----------
    sampling : str | int
        The size of the used grid, either string (e.g. '65x65') or int. The integer will be treated as if obtained from
        zospy.constants.Analysis.SampleSizes_Pow2Plus1_X. Defaults to '65x65'.
    data : str
        The used data type. Should be one of ['TangentialCurvature', 'SagittalCurvature', 'X_Curvature', 'Y_Curvature']
        or int. The integer will be treated as if obtained from zospy.constants.Analysis.SurfaceCurvatureData. Defaults
        to 'TangentialCurvature'.
    remove : str
        Defines whether a reference volume is removed or not. Should be one of ['None', 'BaseROC', 'BestFitSphere'] or
        int. The integer will be treated as if obtained from zospy.constants.Analysis.RemoveOptions. Defaults
        to 'None'.
    surface : int
        The surface that is te be analyzed. defaults to 1.
    show_as : str
        Defines how the data is displayed in OpticStudio. Should be one of ['Surface', 'Contour', 'GreyScale',
        'InverseGreyScale', 'FalseColor', 'InverseFalseColor'] or int. The integer will be treated as if obtained from
        `zospy.constants.Analysis.ShowAs`. Defaults to 'Contour'.
    off_axis_coordinates : bool
        Defines whether apertures defined in the Surface Properties of the surface are considered or not. Defaults to
        `False`.
    contour_format : str
        The contour format. Only usable when showas == 'Contour'. Defaults to ''.
    bfs_criterion : str | int
        The criterion for BFS removal. Only usable when remove == 'BestFitSphere'. Should be one of ['MinimumVolume',
        'MinimumRMS', 'MinimumRMSWithOffset'] or int. The integer will be treated as if obtained from
        constants.Analysis.BestFitSphereOptions. Defaults to 'MinimumVolume'.
    bfs_reverse_direction : bool
        Defines if the sign of the BFS radius should be reversed or not. Only usable when remove == 'BestFitSphere'
        and bfs_criterion == 'MinimumVolume'. Defaults to `False`.
    """

    sampling: str | Annotated[int, Field(ge=0)] = Field(default="65x65", description="Sampling grid size")
    data: ZOSAPIConstant("Analysis.SurfaceCurvatureData") = Field(
        default="TangentialCurvature", description="Data type"
    )
    remove: ZOSAPIConstant("Analysis.RemoveOptions") = Field(default="None_", description="Reference volume removal")
    surface: int = Field(default=1, description="Surface number")
    show_as: ZOSAPIConstant("Analysis.ShowAs") = Field(default="Contour", description="Data display in OpticStudio")
    off_axis_coordinates: bool = Field(default=False, description="Consider apertures defined in surface properties")
    contour_format: str = Field(default="", description="Contour format")
    bfs_criterion: ZOSAPIConstant("Analysis.BestFitSphereOptions") = Field(
        default="MinimumVolume", description="BFS removal criterion"
    )
    bfs_reverse_direction: bool = Field(default=False, description="Reverse BFS radius sign")


class Curvature(BaseAnalysisWrapper[CurvatureResult, CurvatureSettings], analysis_type="SurfaceCurvature"):
    """Surface Curvature analysis."""

    def __init__(
        self,
        *,
        sampling: str | int = "65x65",
        data: str | constants.Analysis.SurfaceCurvatureData = "TangentialCurvature",
        remove: str | constants.Analysis.RemoveOptions = "None_",
        surface: int = 1,
        show_as: str | constants.Analysis.ShowAs = "Contour",
        off_axis_coordinates: bool = False,
        contour_format: str = "",
        bfs_criterion: str | constants.Analysis.BestFitSphereOptions = "MinimumVolume",
        bfs_reverse_direction: bool = False,
    ):
        """Create a new Curvature analysis.

        See Also
        --------
        CurvatureSettings : Settings for the Curvature analysis.
        """
        super().__init__(settings_kws=locals())

    def run_analysis(self) -> list[CurvatureResult] | None:
        """Run the Curvature analysis."""
        self.analysis.Settings.Sampling = getattr(
            constants.Analysis.SampleSizes_Pow2Plus1_X, standardize_sampling(self.settings.sampling)
        )
        self.analysis.Settings.Data = constants.process_constant(
            constants.Analysis.SurfaceCurvatureData, self.settings.data
        )
        self.analysis.Settings.RemoveOption = constants.process_constant(
            constants.Analysis.RemoveOptions, self.settings.remove
        )
        self.analysis.set_surface(self.settings.surface)
        self.analysis.Settings.ShowAs = constants.process_constant(constants.Analysis.ShowAs, self.settings.show_as)
        self.analysis.Settings.ConsiderOffAxisAperture = self.settings.off_axis_coordinates

        if self.analysis.Settings.ShowAs == constants.Analysis.ShowAs.Contour:
            self.analysis.Settings.ContourFormat = self.settings.contour_format

        if self.analysis.Settings.RemoveOption == constants.Analysis.RemoveOptions.BestFitSphere:
            self.analysis.Settings.BestFitSphereOption = constants.process_constant(
                constants.Analysis.BestFitSphereOptions, self.settings.bfs_criterion
            )
            if self.analysis.Settings.BestFitSphereOption == constants.Analysis.BestFitSphereOptions.MinimumVolume:
                self.analysis.Settings.ReverseDirection = self.settings.bfs_reverse_direction

        # Run analysis
        self.analysis.ApplyAndWaitForCompletion()

        return self.get_data_grid()

    def get_data_grid(self) -> CurvatureResult | None:
        """Get the data grids from the Curvature analysis."""
        curvature_description_regex = re.compile(
            rf"Width = (?P<width>\d+(\{config.DECIMAL_POINT}\d+)?), "
            rf"Decenter x = (?P<decenter_x>\d+(\{config.DECIMAL_POINT}\d+)?), "
            rf"y = (?P<decenter_y>\d+(\{config.DECIMAL_POINT}\d+)?) (?P<decenter_unit>\w+)\.",
            re.IGNORECASE,
        )

        if self.analysis.Results.NumberOfDataGrids == 0:
            return None

        if self.analysis.Results.NumberOfDataGrids > 1:
            raise NotImplementedError("Curvature results with multiple data grids are not supported.")

        datagrid = self.analysis.Results.GetDataGrid(0)
        match = curvature_description_regex.match(datagrid.Description)

        if match is None:  # fall back to using exported text files
            logging.warning(
                "Could not obtain description parameters from datagrid.Description, trying to use exported text file"
            )

            self._needs_text_output_file = True
            self._text_output_file, self._remove_text_output_file = self._create_tempfile(None, ".txt")
            match = curvature_description_regex.search(self.get_text_output())

        if match is None:
            raise ValueError(f"Could not parse description: {datagrid.Description}")

        return CurvatureResult(
            width=atox(match.group("width"), float),
            decenter_x=atox(match.group("decenter_x"), float),
            decenter_y=atox(match.group("decenter_y"), float),
            decenter_unit=match.group("decenter_unit"),
            data=unpack_datagrid(datagrid, cell_origin="center"),
        )



"""Zernike Standard Coefficients analysis."""

from __future__ import annotations

from typing import Annotated, TypedDict

from pydantic import Field

from zospy.analyses.base import BaseAnalysisWrapper
from zospy.analyses.decorators import analysis_result, analysis_settings
from zospy.analyses.parsers import ZospyTransformer
from zospy.analyses.parsers.transformers import SimpleField
from zospy.analyses.parsers.types import UnitField  # noqa: TCH001
from zospy.api import constants
from zospy.utils.zputils import standardize_sampling

__all__ = ("ZernikeStandardCoefficients", "ZernikeStandardCoefficientsSettings")


class ZernikeStandardCoefficientsTransformer(ZospyTransformer):
    """Transformer for the output of the Zernike Standard Coefficients analysis."""

    class _ZernikeStandardCoefficient(TypedDict):
        value: float
        formula: str

    def coefficients(self, args):
        """Convert the coefficients to a SimpleField."""
        return SimpleField("Coefficients", dict(args))

    def zernike_coefficient(self, args):
        """Convert a Zernike coefficient to a SimpleField with the value and the formula."""
        index, value, formula = args

        return SimpleField(index, self._ZernikeStandardCoefficient(value=value, formula=formula))


@analysis_result
class ZernikeStandardCoefficient:
    value: float
    formula: str


@analysis_result
class IntegrationData:
    rms_to_chief: UnitField = Field(alias="RMS (to chief)")
    rms_to_centroid: UnitField = Field(alias="RMS (to centroid)")
    variance: UnitField = Field(alias="Variance")
    strehl_ratio: float = Field(alias="Strehl Ratio (Est)")


@analysis_result
class ZernikeStandardCoefficientsResult:
    """Data for the Zernike Standard Coefficients analysis."""

    subaperture_decenter_sx: float | None = Field(alias="Subaperture decenter Sx", default=None)
    subaperture_decenter_sy: float | None = Field(alias="Subaperture decenter Sy", default=None)
    subaperture_radius_sr: float | None = Field(alias="Subaperture radius Sr", default=None)
    surface: str | int = Field(alias="Surface")
    field: UnitField = Field(alias="Field")
    wavelength: UnitField = Field(alias="Wavelength")
    peak_to_valley_to_chief: UnitField = Field(alias="Peak to Valley (to chief)")
    peak_to_valley_to_centroid: UnitField = Field(alias="Peak to Valley (to centroid)")

    from_integration_of_the_rays: IntegrationData = Field(alias="From integration of the rays")
    from_integration_of_the_fitted_coefficients: IntegrationData = Field(
        alias="From integration of the fitted coefficients"
    )

    rms_fit_error: UnitField = Field(alias="RMS fit error")
    maximum_fit_error: UnitField = Field(alias="Maximum fit error")

    coefficients: dict[int, ZernikeStandardCoefficient] = Field(alias="Coefficients")


@analysis_settings
class ZernikeStandardCoefficientsSettings:
    """Settings for the Zernike Standard Coefficients analysis."""

    sampling: str = Field(default="64x64", description="Sampling grid size")
    maximum_term: int = Field(default=37, ge=0, description="Maximum term")
    wavelength: str | Annotated[int, Field(ge=0)] = Field(default=1, description="Wavelength")
    field: str | Annotated[int, Field(ge=0)] = Field(default=1, description="Field")
    reference_opd_to_vertex: bool = Field(default=False, description="Reference OPD to vertex")
    surface: str | Annotated[int, Field(ge=0)] = Field(default="Image", description="Surface")
    sx: float = Field(default=0.0, description="Sx")
    sy: float = Field(default=0.0, description="Sy")
    sr: float = Field(default=0.0, description="Sr")


class ZernikeStandardCoefficients(
    BaseAnalysisWrapper[ZernikeStandardCoefficientsResult, ZernikeStandardCoefficientsSettings],
    analysis_type="ZernikeStandardCoefficients",
    needs_text_output_file=True,
):
    """Zernike Standard Coefficients analysis."""

    def __init__(
        self,
        *,
        sampling: str = "64x64",
        maximum_term: int = 37,
        wavelength: str | int = 1,
        field: str | int = 1,
        reference_opd_to_vertex: bool = False,
        surface: str | int = "Image",
        sx: float = 0.0,
        sy: float = 0.0,
        sr: float = 0.0,
    ):
        super().__init__(settings_kws=locals())

    def run_analysis(self) -> ZernikeStandardCoefficientsResult:
        """Run the Zernike Standard Coefficients analysis."""
        self.analysis.Settings.SampleSize = getattr(
            constants.Analysis.SampleSizes, standardize_sampling(self.settings.sampling)
        )
        self.analysis.Settings.MaximumNumberOfTerms = self.settings.maximum_term
        self.analysis.set_wavelength(self.settings.wavelength)
        self.analysis.set_field(self.settings.field)
        self.analysis.Settings.ReferenceOBDToVertex = (
            self.settings.reference_opd_to_vertex
        )  # TODO: Monitor name with zemax updates
        self.analysis.set_surface(self.settings.surface)
        self.analysis.Settings.Sx = self.settings.sx
        self.analysis.Settings.Sy = self.settings.sy
        self.analysis.Settings.Sr = self.settings.sr

        # Run analysis
        self.analysis.ApplyAndWaitForCompletion()

        # Get results
        return self.parse_output(
            "zernike_standard_coefficients",
            transformer=ZernikeStandardCoefficientsTransformer,
            result_type=ZernikeStandardCoefficientsResult,
        )



"""This file provides autocompletions for the ZOS-API and was automatically generated.
It should not be edited manually.
"""

from __future__ import annotations

__all__ = ("EncircledEnergyTypes", "ExtendedSourceTypes", "GeometricLineEdgeSpreadTypes", "ReferToTypes")

class EncircledEnergyTypes:
    Encircled: EncircledEnergyTypes = None
    X_Only: EncircledEnergyTypes = None
    Y_Only: EncircledEnergyTypes = None
    Ensquared: EncircledEnergyTypes = None

class ExtendedSourceTypes:
    Encircled: ExtendedSourceTypes = None
    X_Only: ExtendedSourceTypes = None
    Y_Only: ExtendedSourceTypes = None
    Ensquared: ExtendedSourceTypes = None
    X_Distrib: ExtendedSourceTypes = None
    Y_Distrib: ExtendedSourceTypes = None

class GeometricLineEdgeSpreadTypes:
    LineEdge: GeometricLineEdgeSpreadTypes = None
    Line: GeometricLineEdgeSpreadTypes = None
    Edge: GeometricLineEdgeSpreadTypes = None

class ReferToTypes:
    ChiefRay: ReferToTypes = None
    Centroid: ReferToTypes = None
    Vertex: ReferToTypes = None
